/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2022 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    "data_preproc_fstec.h"

#ifdef FSTEC_VER_CONV_DEBUGGING
#include    "data_preproc_fstec_debug.h"
#endif

//-----------------------------------------------------------------------------

typedef pair<bool, regex> TMarkedRegex;

//-----------------------------------------------------------------------------

// 'software' fields of the FSTEC list
const vector<int> fstec_vul_soft = {
    FSTEC_VUL_SOFT_VENDOR,
    FSTEC_VUL_SOFT_NAME,
    FSTEC_VUL_SOFT_VER,
    FSTEC_VUL_SOFT_TYPE
    };
// 'environment' fields of FSTEC list
const vector<int> fstec_vul_env = {
    FSTEC_VUL_ENV_VENDOR,
    FSTEC_VUL_ENV_NAME,
    FSTEC_VUL_ENV_VER,
    FSTEC_VUL_ENV_PLATFORM
    };
// Map of indexes of FSTEC fields (compressed to uncompressed)
const map<int, int> fstec_comp_to_uncomp = {
    {FSTEC_VUL_COMP_NAME, FSTEC_VUL_NAME},
    {FSTEC_VUL_COMP_DESCRIPTION, FSTEC_VUL_DESCRIPTION},
    {FSTEC_VUL_COMP_CWE, FSTEC_VUL_CWE},
    {FSTEC_VUL_COMP_IDENTIFY_DATE, FSTEC_VUL_IDENTIFY_DATE},
    {FSTEC_VUL_COMP_CVSS2_SCORE, FSTEC_VUL_CVSS2_SCORE},
    {FSTEC_VUL_COMP_CVSS3_SCORE, FSTEC_VUL_CVSS3_SCORE},
    {FSTEC_VUL_COMP_SEVERITY, FSTEC_VUL_SEVERITY},
    {FSTEC_VUL_COMP_SOLUTION, FSTEC_VUL_SOLUTION},
    {FSTEC_VUL_COMP_STATUS, FSTEC_VUL_STATUS},
    {FSTEC_VUL_COMP_EXPLOIT_STATUS, FSTEC_VUL_EXPLOIT_STATUS},
    {FSTEC_VUL_COMP_FIX_STATUS, FSTEC_VUL_FIX_STATUS},
    {FSTEC_VUL_COMP_CVE_IDS, FSTEC_VUL_OTHER_IDS},
    };

//-----------------------------------------------------------------------------
// Convert version(s) or range of versions of a vulnerable product, taken from
// a FSTEC software list, to the cve-manager format

const regex re_fstec_ver_multiple("([0-9A-Za-z][^\\s]*\\s*)+");
const regex re_fstec_ver_range("\\s*"
    "((О|о)т|(С|с))?\\s*"              // 1-3
    "([0-9A-Za-z][^\\s]*\\s?)*"        // 4
    "((П|п)о|(Д|д)о)?\\s*"             // 5-7
    "([0-9A-Za-z][^\\s]*\\s?)*"        // 8
    "(\\s*\\(?\\s*(В|в)кл.*\\)?\\s*)?" // 9-11
    "$");

pair<pair<vector<string>, int>, int> ConvertFSTECVer(string ver){

    smatch m;
    Trim(ver);

    // If this is a range of versions
    if(regex_match(ver, m, re_fstec_ver_range) &&
            (m[1].length() > 0 || m[5].length() > 0)){

        ver.clear();

        // Left limit
        if(m[1].length() > 0){
            if(m[4].length() == 0){
                return {{{ver}, -1}, 0};
                }
            ver += "[" + m[4].str();
            if(isspace(ver.back()) != 0){
                ver.pop_back();
                }
            ver += ";";
            }
        else{
            if(m[4].length() != 0){
                return {{{ver}, -1}, 1};
                }
            }

        // Rigth limit
        if(m[5].length() > 0){
            if(m[8].length() == 0){
                return {{{ver}, -1}, 2};
                }
            if(ver.empty()){
                ver += "(;";
                }
            ver += m[8].str();
            if(isspace(ver.back()) != 0){
                ver.pop_back();
                }
            }
        else{
            if(m[8].length() != 0){
                return {{{ver}, -1}, 3};
                }
            }

        // Including or not
        if(m[9].length() > 0){
            if(m[5].length() == 0){
                return {{{ver}, -1}, 4};
                }
            ver += "]";
            }
        else{
            ver += ")";
            }
        return {{{ver}, 0}, 5};
        }

    // If this is a list of versions
    if(regex_match(ver, m, re_fstec_ver_multiple)){
        const auto versions = Split(ver);
        return {{versions, versions.size() == 1 ? 1 : 2}, 6};
        }

    return {{{ver}, ver != "-" ? -1 : 3}, 7};
    }

//-----------------------------------------------------------------------------
// Produce compressed item out of various items of a given FSTEC entry

const regex re_disallowed(R"((\,|:))");
const regex re_unwanted(R"(\.(?=(\s|$)))");
const regex fstec_excluded_software_types(
    "("
    "((О|о)перационная\\s*(С|с)истема)|"
    "((М|м)икропрограммный\\s*(К|к)од)|"
    "((С|с)етевое\\s*(С|с)редство)|"
    "((А|а)ппарат)|"
    "(АСУ)"
    ")");

#ifdef FSTEC_VER_CONV_DEBUGGING
map<int, int> count_status = {{-1, 0}, {0, 0}, {1, 0}, {2, 0}, {3, 0}};
#endif

pair<vector<string>, bool> CompressFSTECItem(
        const tree::TSelectedDataEntry &entry,
        const vector<int> &indexes, const size_t col_n,
        const vector<pair<TMarkedRegex, TMarkedRegex>> &re){

    string comp_item;
    vector<string> comp_items;
    vector<TMarkedRegex> to_check_next;

    #ifdef FSTEC_VER_CONV_DEBUGGING
    map<int, int> count_local_status = {{-1, 0}, {0, 0}, {1, 0}, {2, 0}, {3, 0}};
    set<pair<string, int>> not_matched_versions;
    #endif

    // Remove commas and colons only (which are disallowed) or also dots (which
    // have no meaning in vendor names and product names)
    auto remove_unwanted_symbols = [](const string &s,
            const bool is_vendor_or_product){
        const string buf = regex_replace(s, re_disallowed, "");
        return is_vendor_or_product ? regex_replace(buf, re_unwanted, "") : buf;
        };
    // - nested func

    for(const auto &i : indexes){
        const auto &item = entry.at(i);
        if(col_n >= item.size()){
            if(col_n > 0){
                return {{}, false};
                }
            return {{}, true};
            }
        const auto &el = item.at(col_n);
        bool is_vendor_or_product = false;
        if(i == FSTEC_VUL_SOFT_VENDOR || i == FSTEC_VUL_ENV_VENDOR){
            is_vendor_or_product = true;
            for(const auto &[vendor, names] : re){
                const auto &[any_vendor, vendor_re] = vendor;
                if(any_vendor || regex_search(el, vendor_re)){
                    to_check_next.push_back(names);
                    }
                }
            }
        else if(i == FSTEC_VUL_SOFT_NAME || i == FSTEC_VUL_ENV_NAME){
            is_vendor_or_product = true;
            for(const auto &names : to_check_next){
                const auto &[any_name, names_re] = names;
                if(any_name || regex_search(el, names_re)){
                    return {{}, true};
                    }
                }
            }
        else if(i == FSTEC_VUL_SOFT_VER){
            const auto ret = ConvertFSTECVer(el);
            const auto [versions, status] = ret.first;
            #ifdef FSTEC_VER_CONV_DEBUGGING
            count_local_status[status]++;
            if(status < 0){
                not_matched_versions.insert({el, ret.second});
                }
            #endif
            for(const auto &ver : versions){
                comp_items.push_back(comp_item + ":" +
                    (ver.size() <= 64 ? ver : ""));
                }
            continue;
            }
        else if(i == FSTEC_VUL_SOFT_TYPE){
            if(regex_search(el, fstec_excluded_software_types)){
                return {{}, true};
                }
            }
        if(comp_items.empty()){
            comp_item += (comp_item.empty() ? "" : ":" ) +
                remove_unwanted_symbols(el, is_vendor_or_product);
            }
        else{
            for(auto &ci : comp_items){
                ci += ":" + remove_unwanted_symbols(el, is_vendor_or_product);
                }
            }
        }

    #ifdef FSTEC_VER_CONV_DEBUGGING
    for(const auto &[k, v] : count_local_status){
        count_status[k] += v;
        }
    // Print FSTEC versions that the ConvertFSTECVer function failed to convert
    PrintConvVersions(not_matched_versions);
    #endif

    if(comp_items.empty()){
        comp_items.push_back(comp_item);
        }

    return {comp_items, true};
    }

//----------------------------------------------------------------------------
// Replace various types of newline symbols with tabs in a given entry

void ReplaceNewlinesWithTabs(tree::TSelectedDataEntry &entry){

    for(auto &item : entry){
        for(auto &column : item){
            for(;;){
                const auto i = column.find("\r");
                if(i == string::npos){
                    break;
                    }
                column.erase(i, 1);
                }
            for(;;){
                const auto i = column.find("\n");
                if(i == string::npos){
                    break;
                    }
                column.replace(i, 1, "\t");
                }
            }
        }
    }

//-----------------------------------------------------------------------------
// Convert given date-string from the FSTEC format (dd.mm.yyyy) to the
// "yyyy-mm-dd" format

pair<string, bool> ConvertFSTECDate(const string &date){

    const auto l = Split(date, ".");

    if(l.size() != 3){
        return {{}, false};
        }

    return {l[2] + "-" + l[1] + "-" + l[0], true};
    }

//-----------------------------------------------------------------------------
// Replace all occurrences of a substring "target" with a string "sub"

string ReplaceAll(const string &s, const string &target, const string &sub){

    string ret = s;

    for(;;){
        const auto it = ret.find(target);
        if(it == string::npos){
            break;
            }
        ret.replace(it, target.size(), sub);
        }

    return ret;
    }

//-----------------------------------------------------------------------------
// Generate regular expressions for processing products from the FSTEC list

#define POSSIBLE_DELIM "[\\s\\-_]?"

vector<pair<TMarkedRegex, TMarkedRegex>> FormRegex(
        const TExcludedProducts &excluded_products){

    map<string, string> patterns;

    // Compile given regular expression
    auto comp_pattern = [](const string &pattern){
        return TMarkedRegex(
            pattern.empty(), regex(pattern, regex_constants::icase));
        };
    // - nested func

    for(const auto &[vendor, names] : excluded_products){
        string vendor_pattern, name_pattern;
        if(vendor != "*"){
            vendor_pattern =
                "^(" + ReplaceAll(vendor, "~", POSSIBLE_DELIM) + ")";
            }
        string names_pattern;
        if(!names.second.empty()){
             for(const auto &name : names.second){
                names_pattern += (names_pattern.empty() ? "(" : "|") +
                    ReplaceAll(name, "~", POSSIBLE_DELIM);
                }
            if(!names_pattern.empty()){
                names_pattern = "^(?!" + names_pattern + ")";
                }
            }
        else{
            for(const auto &name : names.first){
                if(name == "*"){
                    names_pattern.clear();
                    break;
                    }
                names_pattern += (names_pattern.empty() ? "^(" : "|") +
                    ReplaceAll(name, "~", POSSIBLE_DELIM);
                }
            }
        if(!names_pattern.empty()){
            names_pattern += ")";
            }
        patterns[vendor_pattern] = names_pattern;
        }

    vector<pair<TMarkedRegex, TMarkedRegex>> re;

    for(const auto &[vendor_pattern, names_pattern] : patterns){
        re.push_back(
            {comp_pattern(vendor_pattern), comp_pattern(names_pattern)});
        }

    return re;
    }

#undef POSSIBLE_DELIM

//-----------------------------------------------------------------------------
// Compress items of a software list or an environments list

bool CompressFSTECSoftOrEnv(const unsigned fstec_vul_comp_software_or_env_index,
        tree::TSelectedDataEntry &entry, tree::TSelectedDataEntry &entry0,
        const vector<pair<TMarkedRegex, TMarkedRegex>> &re){

    const auto &fstec_vul_soft_or_env =
        fstec_vul_comp_software_or_env_index == FSTEC_VUL_COMP_SOFTWARE_LIST ?
        fstec_vul_soft : fstec_vul_env;

    for(size_t i = 0;; i++){
        auto [software_or_env_list, ok] =
            CompressFSTECItem(entry0, fstec_vul_soft_or_env, i, re);
        if(!ok){
            break;
            }
        if(software_or_env_list.empty()){
            continue;
            }
        for(auto &software_or_env : software_or_env_list){
            software_or_env = regex_replace(software_or_env, whitespace, "_");
            entry[fstec_vul_comp_software_or_env_index].push_back(software_or_env);
            }
        }

    if(entry[fstec_vul_comp_software_or_env_index].empty()){
        entry0.clear();
        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Compress sets of FSTEC-VUL fields into fewer quantity of sets

bool PreprocessFSTECVul(tree::TSelectedData &selected_data,
        const TExcludedProducts &excluded_products, string &msg){

    tree::TSelectedDataEntry entry(FSTEC_VUL_COMP_Q);
    const regex non_num_begin(R"(^[^0-9]+)");
    const auto re = FormRegex(excluded_products);

    // For all entries
    try{
        for(auto &entry0 : selected_data){

            // Splitting vul id into year and number
            auto [vul_id, ok] = GetVulId(entry0);
            if(!ok){
                entry0.clear();
                continue;
                }
            entry.at(VUL_YEAR).swap(vul_id.y);
            entry.at(VUL_NUM).swap(vul_id.n);

            // Compressing soft fields into
            // {vendor1:name1:ver1:type1, vendor2:name2:ver2:type2, ... }
            if(!CompressFSTECSoftOrEnv(FSTEC_VUL_COMP_SOFTWARE_LIST,
                    entry, entry0, re)){
                continue;
                }

            // Compressing env fields into
            // {vendor1:name1:ver1:platform1, vendor2:name2:ver2:platform2, ... }
            if(!CompressFSTECSoftOrEnv(FSTEC_VUL_COMP_ENV,
                    entry, entry0, re)){
                continue;
                }

            string sources;

            // In the FSTEC list sources are stored as a single string with the
            // newline symbol as a delimiter
            if(entry0.size() >= FSTEC_VUL_SOURCES &&
                    !entry0.at(FSTEC_VUL_SOURCES).empty()){
                sources = entry0.at(FSTEC_VUL_SOURCES).front();
                }

            // Splitting the "sources" string and pushing the elements into the
            // entry one by one
            for(auto &source: Split(sources)){
                Trim(source);
                entry[FSTEC_VUL_COMP_SOURCES].push_back(source);
                }

            // Pushing data that don't need to be compressed
            for(auto &[i_comp, i_not_comp] : fstec_comp_to_uncomp){
                auto &items = entry0.at(i_not_comp);
                if(items.empty()){
                    continue;
                    }
                if(i_comp == FSTEC_VUL_COMP_CVE_IDS){
                    for(auto &other_id : items){
                        Trim(other_id);
                        if(other_id.compare(0, 4, "CVE-") == 0){
                            // There are CVE-CVE-yyyy-nnnn records in the FSTEC list
                            const auto cve_id =
                                regex_replace(other_id, non_num_begin, "");
                            entry[i_comp].push_back(cve_id);
                            }
                        }
                    continue;
                    }
                const auto &item = items.front();
                // Dates must be converted from the "dd.mm.yyyy" format to the
                // "yyyy-mm-dd" format
                if(i_comp == FSTEC_VUL_COMP_IDENTIFY_DATE){
                    const auto [conv_date, ok] = ConvertFSTECDate(item);
                    if(!ok){
                        continue;
                        }
                    entry[i_comp].push_back(conv_date);
                    }
                else{
                    entry[i_comp].push_back(item);
                    }
                }

            // Replacing "\r\n" and "\n" with "\t" in all items
            ReplaceNewlinesWithTabs(entry);

            // Replacing the element with formed entry
            entry0.swap(entry);
            // Rectreating a clear entry
            entry.Reset(FSTEC_VUL_COMP_Q);
            }
        }
    catch(const exception &e){
        msg = ErrMsg("Unable to preprocess FSTEC vul selected data", e);
        return false;
        }

    #ifdef FSTEC_VER_CONV_DEBUGGING
    PrintCountStatus(count_status);
    #endif

    return true;
    }

//-----------------------------------------------------------------------------
