/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2022 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    <curl/curl.h>

#include    "data_preproc_nvd.h"
#include    "data_preproc_nvd_software_interrelation.h"

//-----------------------------------------------------------------------------

// Components of a CPE name
const vector<int> cpe_components = {
    CPE_SEP_PART,
    CPE_SEP_VENDOR,
    CPE_SEP_PRODUCT,
    CPE_SEP_VER,
    CPE_SEP_REL,
    CPE_SEP_AFTER_REL
    };

//-----------------------------------------------------------------------------
// Form a message about excluded CPEs

string ExtraMsg(const int accepted, const int excl_by_part,
        const int excl_by_name, const int excl_entries=0){

    const int total = accepted + excl_by_part + excl_by_name;

    auto prc = [&](const int piece){
        return 100 * piece / total;
        };

    ostringstream sstr;
    if(excl_entries > 0){
        sstr << excl_entries << "(" << prc(excl_entries) << "%) "
             << "entries excl, ";
        }
    sstr << excl_by_part << " (" << prc(excl_by_part) << "%) "
         << "CPE(s) excl by part and "
         << excl_by_name << " (" << prc(excl_by_name) << "%) "
         << "CPE(s) excl by name";
    if(excl_entries > 0){
        sstr << " (in remaining entries)";
        }

    return sstr.str();
    }

//-----------------------------------------------------------------------------
// Check if given combination of vendor and product is excluded

bool IsExcludedCPE(const string &vendor, const string &product,
        const TExcludedProducts &excluded_cpe){

    auto it = excluded_cpe.find(vendor);

    if(it == excluded_cpe.end()){
        it = excluded_cpe.find("*");
        if(it == excluded_cpe.end()){
            return false;
            }
        }

    const auto &products = it->second;
    const auto &[excluded_products, exceptional_products] = products;

    const bool is_excluded =
        (excluded_products.find(product) != excluded_products.end()) ||
        (excluded_products.find("*") != excluded_products.end());

    const bool is_exception =
        (exceptional_products.find(product) != exceptional_products.end());

    return is_excluded && !is_exception;
    }

//----------------------------------------------------------------------------
// Check if CPE is of 'application' part

bool AppPart(const string &cpe){

    return (cpe.size() > CPE_PART_POS) &&
        (cpe.at(CPE_PART_POS) == 'a');
    }

//----------------------------------------------------------------------------
// Check if CPE is of 'operating system' part and of 'linux' vendor

bool LinuxOS(const string &cpe){
    return (cpe.size() > CPE_LINUX_PRODUCT_POS) &&
        (cpe.at(CPE_PART_POS) == 'o') &&
        (cpe.compare(CPE_VENDOR_POS, 6, "linux:") == 0);
    }

//----------------------------------------------------------------------------
// Check that specified position of a delimiter between components of a given
// CPE is valid

bool ValidPosOfCPECompDelim(const size_t pos, const string &cpe){

    return (pos != string::npos) &&
        (pos > 0) &&
        (cpe.size() > (pos + 1)) &&
        (cpe.at(pos + 1) != ':');
    }

//----------------------------------------------------------------------------
// Unescape URL codes in a given product name

void UnescapeURLCodes(string &product){

    string product0 = product;

    CURL *curl = curl_easy_init();

    do{
        product0 = product;
        if(!product.empty() && curl){
            int output_length;
            char *output = curl_easy_unescape(curl, product.c_str(),
                product.size(), &output_length);
            if(output){
                product.assign(output, output_length);
                curl_free(output);
                }
            }
        } while(product != product0);

    curl_easy_cleanup(curl);
    }

//----------------------------------------------------------------------------
// Split full CPE name into five components

/*
    CPE format version 2.3 is:
        cpe:2.3:<part>:<vendor>:<product>:<version>:<release>:<rest>
        01234567     89
    for example, "SomeVedor CoolProg 1.2.3 Beta" could look like this:
        cpe:2.3:a:some_vendor:cool_prog:1.2.3:beta
    CPE parts:
        h  =  hardware part
        o  =  operating system part
        a  =  application part
*/

map<int, string> SplitCPE(const string &cpe){

    map<int, string> splitted_cpe;
    set<int> esc_pos;

    for(size_t i = 0, pos = CPE_PART_POS - 2, pos0 = pos;
            i < cpe_components.size();
            i++){

        pos = cpe.size() > (pos + 1) ? cpe.find(CPE_DELIMITER, pos + 1) : (-1);

        if(!ValidPosOfCPECompDelim(pos, cpe)){
            return splitted_cpe;
            }

        // If delimiter-symbol is escaped
        if(cpe[pos - 1] == '\\'){
            esc_pos.insert(pos - 1);
            i--;
            continue;
            }

        if(i > 0){
            auto &component = splitted_cpe[cpe_components.at(i - 1)];
            // If there was at least one escaped delimiter-symbol on the way
            if(!esc_pos.empty()){
                for(size_t j = (pos0 + 1); j < pos; j++){
                    if(esc_pos.find(j) != esc_pos.end()){
                        if(component.empty() || component.back() != '_'){
                            component += '_';
                            }
                        j++;
                        continue;
                        }
                    component += cpe[j];
                    }
                esc_pos.clear();
                }
            else{
                const size_t len = pos - pos0 - 1;
                component = cpe.substr(pos0 + 1, len);
                }
            }

        pos0 = pos;
        }

    // Unescaping URL codes in product names
    string &product = splitted_cpe[CPE_SEP_PRODUCT];
    UnescapeURLCodes(product);

    // Making sure there are no commas in versions
    string &ver = splitted_cpe[CPE_SEP_VER];
    replace(ver.begin(), ver.end(), ',', '.');

    return splitted_cpe;
    }

//-----------------------------------------------------------------------------
// Throw away false positive vulnerabilities and define a range of vulnerable
// versions for a software list of a given entry

void ProcessSoftwareList(tree::TSelectedDataEntry &entry,
        const TExcludedProducts &excluded_cpe, int &count_excl_cpe_by_part,
        int &count_excl_cpe_by_name){

    tree::TSelectedDataEntryItem software_item;

    map<int, string> indexes_of_cpe_with_unknown_ver;
    set<string> cpe_with_specified_ver;
    auto &software_list = entry.at(NVD_CVE_SOFTWARE);

    for(size_t i = 0; i < software_list.size(); i++){

        const auto &software_is_vulnerable =
            entry.at(NVD_CVE_SOFTWARE_VULNERABLE);

        if(software_is_vulnerable.size() <= i){
            software_item.push_back("");
            continue;
            }

        const string &cpe = software_list.at(i);

        // We need CPEs of 'application' part only
        if(!AppPart(cpe) && !LinuxOS(cpe)){
            software_item.push_back("");
            count_excl_cpe_by_part++;
            continue;
            }

        // Splitting this CPE
        const map<int, string> splitted_cpe = SplitCPE(cpe);
        if(splitted_cpe.size() < 5){
            software_item.push_back("");
            continue;
            }

        const string &vendor = splitted_cpe.at(CPE_SEP_VENDOR);
        const string &product = splitted_cpe.at(CPE_SEP_PRODUCT);
        const string &ver = splitted_cpe.at(CPE_SEP_VER);
        const string &rel = splitted_cpe.at(CPE_SEP_REL);

        if(vendor.empty() || product.empty()){
            software_item.push_back("");
            continue;
            }

        // Excluding some of the CPEs
        if(IsExcludedCPE(vendor, product, excluded_cpe)){
            software_item.push_back("");
            count_excl_cpe_by_name++;
            continue;
            }

        string joined_cpe = (software_is_vulnerable.at(i) == "1" ? "" : "~") +
            vendor + ":" + product + ":";
        bool got_specified_ver = false;

        // Forming version component (single ver or range of versions)
        if(ver == "*" || ver.empty()){

            // There are incidents of presence of misleading ':' chars -- removing leading
            // and trailing ':' chars and replacing the rest of them with the '.' char
            for(size_t j = NVD_CVE_SOFTWARE_START_EXCL; j <= NVD_CVE_SOFTWARE_END_INCL; j++){
                auto &item = entry.at(j);
                if(item.size() <= i){
                    continue;
                    }
                auto &el = item.at(i);
                Trim(el, ":");
                replace_if(el.begin(), el.end(), [](char c){return c == ':';}, '.');
                }

            string range = "";
            const auto &start_excl = entry.at(NVD_CVE_SOFTWARE_START_EXCL);
            const auto &start_incl = entry.at(NVD_CVE_SOFTWARE_START_INCL);
            const auto &end_excl = entry.at(NVD_CVE_SOFTWARE_END_EXCL);
            const auto &end_incl = entry.at(NVD_CVE_SOFTWARE_END_INCL);
            if(start_excl.size() > i && !start_excl.at(i).empty()){
                range = "(" + start_excl.at(i);
                got_specified_ver = true;
                }
            else if(start_incl.size() > i && !start_incl.at(i).empty()){
                range = "[" + start_incl.at(i);
                got_specified_ver = true;
                }
            else{
                range = "(";
                }
            range += ";";
            if(end_excl.size() > i && !end_excl.at(i).empty()){
                range += end_excl.at(i) + ")";
                got_specified_ver = true;
                }
            else if(end_incl.size() > i && !end_incl.at(i).empty()){
                range += end_incl.at(i) + "]";
                got_specified_ver = true;
                }
            else{
                range += ")";
                }
            joined_cpe += range;
            }
        else{
            got_specified_ver = ver != "-";
            joined_cpe += ver;
            }

        if(got_specified_ver){
            cpe_with_specified_ver.insert(product);
            }
        else{
            indexes_of_cpe_with_unknown_ver[software_item.size()] = product;
            }

        joined_cpe += ":" + rel;
        software_item.push_back(joined_cpe);
        }

    // Removing CPE with unknown ver if there is a CPE with specified ver
    if(!cpe_with_specified_ver.empty() &&
            !indexes_of_cpe_with_unknown_ver.empty()){

        for(size_t i = 0; i < software_item.size(); i++){
            const auto indexes_it = indexes_of_cpe_with_unknown_ver.find(i);
            // If this CPE does not have unspecified version
            if(indexes_it == indexes_of_cpe_with_unknown_ver.end()){
                continue;
                }
            const auto cpe_it = cpe_with_specified_ver.find(indexes_it->second);
            // If there is a CPE with specifiend version for the same product
            if(cpe_it != cpe_with_specified_ver.end()){
                software_item[i] = "";
                }
            }
        }

    // Forming a map of indexes of binded vulnerable products
    const auto [stages, stages_ready] =
        FormStagesOfOperations(entry, software_item.size());
    if(!stages_ready){
        entry.clear();
        return;
        }

    #ifdef DATA_PREPROCESS_DEBUGGING
    PrintStages(stages);
    #endif

    // Form a disjunction of conjunctions of indexes of vulnerable products
    auto [disjunction_of_conjunctions, disjunction_of_conjunctions_ready] =
        FormDisjunctionOfConjunctions(stages);
    if(!disjunction_of_conjunctions_ready){
        entry.clear();
        return;
        }

    // Filtering a disjunction of conjunctions of indexes of vulnerable products
    FilterDisjunctionOfConjunctions(disjunction_of_conjunctions, software_item);

    // Reconstructing a software item
    ReconstructSoftwareItem(disjunction_of_conjunctions, software_item);

    // Excluding CVE records with empty list of vulnerable software
    // by clearing the entry
    if(software_item.empty()){
        entry.clear();
        }
    else{
        software_list.swap(software_item);
        }
    }

//-----------------------------------------------------------------------------
// Adjust NVD vulnerabilities data that was selected from a JSON document

bool PreprocessNVDVul(tree::TSelectedData &selected_data,
        const TExcludedProducts &excluded_cpe, string &msg){

    int count_excl_cpe_by_part = 0;
    int count_excl_cpe_by_name = 0;
    int count_excl_entries = 0;
    int count_accepted = 0;

    try{

        for(auto &entry : selected_data){

            const int count_excl_cpe_by_part0 = count_excl_cpe_by_part;
            const int count_excl_cpe_by_name0 = count_excl_cpe_by_name;

            // Excluding CVE records with insufficient number of columns
            if(entry.size() != NVD_CVE_Q){
                entry.clear();
                continue;
                }

            // Splitting vul id into year and number
            auto [vul_id, ok] = GetVulId(entry);
            if(!ok){
                entry.clear();
                continue;
                }
            entry.at(VUL_YEAR).swap(vul_id.y);
            entry.at(VUL_NUM).swap(vul_id.n);

            #ifdef DATA_PREPROCESS_DEBUGGING
            cout << endl << "Entry before 'ProcessSoftwareList': '"
                 << entry.Stringify(SEL_L_BORDER, SEL_R_BORDER, SEL_DELIM)
                 << "'" << endl;
            #endif

            // Process the "software list" item of this entry (if software list
            // content no useful information the entry will be cleared)
            ProcessSoftwareList(entry, excluded_cpe, count_excl_cpe_by_part,
                count_excl_cpe_by_name);

            #ifdef DATA_PREPROCESS_DEBUGGING
            cout << "Entry after 'ProcessSoftwareList': '"
                 << entry.Stringify(SEL_L_BORDER, SEL_R_BORDER, SEL_DELIM)
                 << "'" << endl;
            #endif

            if(entry.empty()){
                if(count_excl_cpe_by_part != count_excl_cpe_by_part0 ||
                        count_excl_cpe_by_name != count_excl_cpe_by_name0){
                    count_excl_cpe_by_part = count_excl_cpe_by_part0;
                    count_excl_cpe_by_name = count_excl_cpe_by_name0;
                    count_excl_entries++;
                    }
                continue;
                }

            count_accepted++;

            // Adding refs tags
            if(entry.at(NVD_CVE_REFS).size() !=
                    entry.at(NVD_CVE_REFS_TAG).size()){
                continue;
                }

            tree::TSelectedDataEntryItem refs_item;
            tree::TSelectedDataEntryItem patch_item;

            for(size_t i = 0; i < entry.at(NVD_CVE_REFS_TAG).size(); i++){
                string &tag = entry.at(NVD_CVE_REFS_TAG).at(i);
                Trim(tag, "{ }");
                const string &ref = entry.at(NVD_CVE_REFS).at(i);
                if(tag.empty()){
                    refs_item.push_back(ref);
                    continue;
                    }
                tag = ToLower(tag);
                if(tag.find("patch") != string::npos){
                    patch_item.push_back(ref);
                    }
                else{
                    refs_item.push_back(
                        "#" + regex_replace(tag, whitespace, "-") + "#" + ref);
                    }
                }

            entry.at(NVD_CVE_REFS).swap(refs_item);
            entry.at(NVD_CVE_PATCH).swap(patch_item);

            // Clearing unneeded fields (they will be cutted out automatically
            // during CSV import because a table to which the import is
            // performed does not have so many columns)
            for(size_t i = NVD_CVE_SUMMARY + 1; i < NVD_CVE_Q; i++){
                entry.at(i).clear();
                }
            }
        }
    catch(const exception &e){
        msg = ErrMsg("Unable to preprocess NVD vul selected data", e);
        return false;
        }

    if(count_excl_cpe_by_part > 0 || count_excl_cpe_by_name > 0){
        msg = ExtraMsg(count_accepted, count_excl_cpe_by_part,
            count_excl_cpe_by_name, count_excl_entries);
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Combine all references of each {part, vendor, product} combination without
// any regards to versions, determine if this product is for non free OS only
// (that's the only thing the "title" field is useful for)

struct CPE_desc{
    set<string> refs;
    int non_free = 0;
    };

bool PreprocessCPEDict(tree::TSelectedData &selected_data,
        const TExcludedProducts &excluded_cpe, string &msg){

    int count_excl_by_part = 0;
    int count_excl_by_name = 0;
    int count_accepted = 0;

    // Regex for checking if a title states that this is an app for non-free OS
    // (case-insensitive)
    const regex for_nonfree_os(
        " (for ((microsoft ?)?windows|mac ?(os)?|android)|windows edition)( |$)",
        regex_constants::icase);

    try{

        map<string, map<string, map<string, CPE_desc>>> cpe_descriptions;

        // For all entries of selected data
        for(auto &sel_data_entry : selected_data){

            if(sel_data_entry.at(CPE_FULLNAME).empty()){
                throw runtime_error("CPE name is missing");
                }

            const string &cpe = sel_data_entry.at(CPE_FULLNAME).front();

            // We need CPEs of 'application' part only
            if(!AppPart(cpe) && !LinuxOS(cpe)){;
                count_excl_by_part++;
                sel_data_entry.clear();
                continue;
                }

            // Splitting a CPE name
            const auto splitted_cpe = SplitCPE(cpe);
            if(splitted_cpe.size() < 5){
                sel_data_entry.clear();
                continue;
                }

            const string &part  = splitted_cpe.at(CPE_SEP_PART);
            const string &vendor = splitted_cpe.at(CPE_SEP_VENDOR);
            const string &product = splitted_cpe.at(CPE_SEP_PRODUCT);

            // Removing incomplete CPEs
            if(part.empty() || vendor.empty() || product.empty()){
                sel_data_entry.clear();
                continue;
                }

            // Excluding some of the CPEs
            if(IsExcludedCPE(vendor, product, excluded_cpe)){
                count_excl_by_name++;
                sel_data_entry.clear();
                continue;
                }

            auto &cpe_desc = cpe_descriptions[part][vendor][product];

            // Adding refs for this CPE
            for(const auto &el : sel_data_entry.at(CPE_REFS)){
                cpe_desc.refs.insert(el);
                }

            // Modifying the "non free OS" flag for this CPE
            if(sel_data_entry.at(CPE_TITLE).empty()){
                throw runtime_error("CPE title is missing");
                }

            const auto &title = sel_data_entry.at(CPE_TITLE).front();
            cpe_desc.non_free |= regex_search(title, for_nonfree_os) ?
                0b10 : 0b01;

            sel_data_entry.clear();
            count_accepted++;
            }

        selected_data.clear();

        // Reforming the selected data using a formed map of CPE descriptions
        for(const auto &[part, vendors] : cpe_descriptions){
            for(const auto &[vendor, products] : vendors){
                for(const auto &[product, desc] : products){

                    tree::TSelectedDataEntry entry(CPE_SEP_Q);

                    entry[CPE_SEP_PART].push_back(part);
                    entry[CPE_SEP_VENDOR].push_back(vendor);
                    entry[CPE_SEP_PRODUCT].push_back(product);
                    for(const auto &ref : desc.refs){
                        entry[CPE_SEP_REFS].push_back(ref);
                        }
                    entry[CPE_SEP_NON_FREE].push_back(to_string(desc.non_free));

                    selected_data.push_back(entry);
                    }
                }
            }
        }
    catch(const exception &e){
        msg = ErrMsg("Unable to preprocess CPE dict selected data", e);
        return false;
        }

    if(count_excl_by_part > 0 || count_excl_by_name > 0){
        msg = ExtraMsg(count_accepted, count_excl_by_part, count_excl_by_name);
        }

    return true;
    }

//-----------------------------------------------------------------------------
