/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2022 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    <regex>

#include    "data_preproc_nvd_software_interrelation.h"

// Regular expressions to validate operators
const regex is_conjunction("^AND[0-9]*$");
const regex is_disjunction("^OR[0-9]*$");

#ifdef DATA_PREPROCESS_DEBUGGING

//-----------------------------------------------------------------------------
// Print stages

void PrintStages(const vector<map<string, set<int>>> &stages,
        const unsigned short indent){

    for(short i = 0; i < indent; i++){
        cout << "\t";
        }

    cout << "stages: {";

    for(auto m_it = stages.rbegin(); m_it != stages.rend(); m_it++){
        if(m_it != stages.rbegin()){
            cout << ", ";
            }
        cout << stages.size() - (m_it - stages.rbegin())
             << ": " << Stringify(*m_it);
        }

    cout << "}" << endl;
    }

//-----------------------------------------------------------------------------
// Print given disjunction of conjunctions

void PrintCombination(vector<set<int>> &comb, const unsigned short indent){

    for(short i = 0; i < indent; i++){
        cout << "\t";
        }

    cout << "combination: {";

    for(auto s = comb.begin(); s != comb.end(); s++){
        if(s != comb.begin()){
            cout << ", ";
            }
        cout << Stringify(*s);
        }

    cout << "}" << endl;
    }

//-----------------------------------------------------------------------------
// Print given map of processed combinations

void PrintProcessedCombinations(map<int, set<int>> &comb,
        const unsigned short indent){

    for(short i = 0; i < indent; i++){
        cout << "\t";
        }

    cout << "processed_combinations: " << Stringify(comb) << endl;
    }

#endif

//-----------------------------------------------------------------------------
// Form a map of indexes of binded vulnerable products (binded by specific
// operators that used as keys)

#define NSTAGES 2

pair<vector<map<string, set<int>>>, bool> FormStagesOfOperations(
        tree::TSelectedDataEntry &entry, size_t software_item_size){

    vector<map<string, set<int>>> stages;

    for(int i = 0; i < NSTAGES; i++){

        const auto &operators = entry.at(NVD_CVE_SOFTWARE_LOGIC1 + i);

        if(operators.size() != software_item_size &&
                (i == 0 || !operators.empty())){
            return {{}, false};
            }

        map<string, set<int>> stage;

        for(auto op = operators.begin(); op != operators.end(); op++){
            if(op->empty()){
                continue;
                }
            const auto n = op - operators.begin();
            stage[*op].insert(n);
            }

        if(!stage.empty()){
            stages.push_back(stage);
            }
        }

    return {stages, true};
    }

#undef NSTAGES

//-----------------------------------------------------------------------------
// Process a group of elements binded by a conjunction operator (AND)

void ProcessConjunction(vector<set<int>> &disjunction_of_conjunctions,
        const set<int> &group, map<int, set<int>> &processed_combinations){

    #ifdef DATA_PREPROCESS_DEBUGGING
    PrintProcessedCombinations(processed_combinations, 2);
    #endif

    // If there are no disjunctions
    if(disjunction_of_conjunctions.empty()){
        disjunction_of_conjunctions.push_back(group);
        for(const auto &el : group){
            processed_combinations[el].insert(group.begin(), group.end());
            }
        return;
        }

    vector<set<int>> new_disjunction_of_conjunctions;

    for(auto &conjunction : disjunction_of_conjunctions){

        #ifdef DATA_PREPROCESS_DEBUGGING
        cout << "\t\tProcessing conjunction "
             << Stringify(conjunction) << endl;
        #endif

        const bool got_intersection = [&](void){
            for(auto &el : group){
                if(conjunction.find(el) != conjunction.end()){
                    return true;
                    }
                }
            return false;
            }();

        // Skipping if this conjunction doesn't contain any elements of this group
        if(!got_intersection){
            #ifdef DATA_PREPROCESS_DEBUGGING
            cout << "\t\t\tThis conjunction doesn't contain any elements "
                 << "of this group" << endl;
            #endif
            new_disjunction_of_conjunctions.push_back(conjunction);
            continue;
            }

        #ifdef DATA_PREPROCESS_DEBUGGING
        cout << "\t\t\tThis conjunction contains some elements of this group" << endl;
        #endif

        vector<set<int>> new_conjunctions;
        vector<set<int>> conjunction_derivatives = {conjunction};

        // This conjunction contains some elements of this group
        for(auto &el1 : group){

            #ifdef DATA_PREPROCESS_DEBUGGING
            cout << "\t\t\tCurrent element of the group: " << el1 << endl;
            #endif

            set<int> new_conjunction;

            for(auto derivative_it = conjunction_derivatives.rbegin();
                    derivative_it != conjunction_derivatives.rend();
                    derivative_it++){

                const bool is_associated = [&](void){
                    const auto it = processed_combinations.find(el1);
                    if(it != processed_combinations.end()){
                        const auto &el1_associates = it->second;
                        for(const auto &el2 : el1_associates){
                            if(derivative_it->find(el2) != derivative_it->end()){
                                return true;
                                }
                            }
                        }
                    return false;
                    }();

                // If el1 is already associated with some elements of this
                // conjunction
                if(is_associated){
                    #ifdef DATA_PREPROCESS_DEBUGGING
                    cout << "\t\t\t\tThis element is already associated "
                        "with some elements of this conjunction" << endl;
                    #endif
                    continue;
                    }

                #ifdef DATA_PREPROCESS_DEBUGGING
                cout << "\t\t\t\tINSERTING THIS ELEMENT" << endl;
                #endif
                new_conjunction = *derivative_it;
                new_conjunction.insert(el1);
                break;
                }

            if(new_conjunction.empty()){
                continue;
                }

            #ifdef DATA_PREPROCESS_DEBUGGING
            cout << "\t\t\t\tNew conjunction: "
                 << Stringify(new_conjunction)
                 << endl;
            #endif

            new_conjunctions.push_back(new_conjunction);
            conjunction_derivatives.push_back(new_conjunction);

            for(const auto &el2 : new_conjunction){
                processed_combinations[el2].insert(el1);
                }
            }

        for(const auto &new_conjunction : new_conjunctions){
            new_disjunction_of_conjunctions.push_back(new_conjunction);
            }

        #ifdef DATA_PREPROCESS_DEBUGGING
        PrintCombination(disjunction_of_conjunctions, 4);
        PrintProcessedCombinations(processed_combinations, 4);
        #endif
        }

    new_disjunction_of_conjunctions.swap(disjunction_of_conjunctions);
    }

//-----------------------------------------------------------------------------
// Process a group of elements binded by a disjunction operator (OR)

void ProcessDisjunction(vector<set<int>> &disjunction_of_conjunctions,
        const set<int> &group, map<int, set<int>> &processed_combinations){

    for(auto &el0 : group){

        const bool not_in_any_conjunctions = [&](void){
            for(const auto &conjunction : disjunction_of_conjunctions){
                if(conjunction.find(el0) != conjunction.end()){
                    return false;
                    }
                }
            return true;
            }();

        if(not_in_any_conjunctions){
            disjunction_of_conjunctions.push_back({el0});
            }

        processed_combinations[el0].insert(group.begin(), group.end());
        }
    }

//-----------------------------------------------------------------------------
// Form a disjunction of conjunctions of indexes of vulnerable products (each
// set is a set of indexes binded with a conjunction operators and a vector
// itself is a disjunction of this kind of sets)

pair<vector<set<int>>, bool> FormDisjunctionOfConjunctions(
        const vector<map<string, set<int>>> &stages){

    vector<set<int>> disjunction_of_conjunctions;
    map<int, set<int>> processed_combinations;

    for(auto stage = stages.rbegin(); stage != stages.rend(); stage++){

        #ifdef DATA_PREPROCESS_DEBUGGING
        cout << "Processing stage #" << (stages.rend() - stage) << endl;
        #endif

        // Processing each stage
        for(auto &[op, group] : *stage){

            #ifdef DATA_PREPROCESS_DEBUGGING
            cout << "\top is '"  << op << "', "
                 << "group is " << Stringify(group) << ", ";
            #endif

            // If this is "AND" operator
            if(regex_search(op, is_conjunction)){
                #ifdef DATA_PREPROCESS_DEBUGGING
                cout << "this is 'AND' operator" << endl;
                #endif
                ProcessConjunction(disjunction_of_conjunctions, group,
                    processed_combinations);
                }

            // If this is "OR" operator
            else if(regex_search(op, is_disjunction)){
                #ifdef DATA_PREPROCESS_DEBUGGING
                cout << "this is 'OR' operator" << endl;
                #endif
                ProcessDisjunction(disjunction_of_conjunctions, group,
                    processed_combinations);
                }

            // This is invalid operator
            else{
                #ifdef DATA_PREPROCESS_DEBUGGING
                cout << "this is unknown operator!" << endl;
                #endif
                return {{}, false};
                }

            #ifdef DATA_PREPROCESS_DEBUGGING
            PrintCombination(disjunction_of_conjunctions, 2);
            #endif
            }
        }

    return {disjunction_of_conjunctions, true};
    }

//-----------------------------------------------------------------------------
// Filtering a given combination of elements of a given software item

void FilterDisjunctionOfConjunctions(
        vector<set<int>> &disjunction_of_conjunctions,
        tree::TSelectedDataEntryItem &software_item){

    vector<set<int>> disjunction_of_conjunctions_filtered;

    for(const auto &conjunction : disjunction_of_conjunctions){

        const bool no_empty_elements = [&](void){
            for(const auto &el : conjunction){
                if(software_item.at(el).empty()){
                    return false;
                    }
                }
            return true;
            }();

        if(no_empty_elements){
            disjunction_of_conjunctions_filtered.push_back(conjunction);
            }
        }

    disjunction_of_conjunctions.swap(disjunction_of_conjunctions_filtered);
    }

//-----------------------------------------------------------------------------
// Reconstruct software item using given combination and current values of
// a software item

void ReconstructSoftwareItem(
        vector<set<int>> &disjunction_of_conjunctions,
        tree::TSelectedDataEntryItem &software_item){

    tree::TSelectedDataEntryItem software_item_new;

    for(const auto &conjunction : disjunction_of_conjunctions){

        if(conjunction.size() == 1){
            auto it = conjunction.begin();
            software_item_new.push_back(software_item.at(*it));
            continue;
            }

        string conjunction_str;

        for(auto it = conjunction.begin(); it != conjunction.end(); it++){
            if(it != conjunction.begin()){
                conjunction_str += "&&";
                }
            conjunction_str += software_item.at(*it);
            }

        software_item_new.push_back(conjunction_str);
        }

    software_item.swap(software_item_new);
    }
