/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2022 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "database.h"
#include "rpm.h"

//-----------------------------------------------------------------------------
// Create DB structure

bool TSpecificDBInterface::CreateFullStruct(const vector<TBranch> &branches,
        const vector<string> &excluded_vulsrc){

    auto excl = [&](const string &target){
        const auto &excl = excluded_vulsrc;
        return find(excl.begin(), excl.end(), target) != excl.end();
        };

    const bool success = ClearDB() &&
        (excl(TARGET_NVD) || CreateCPETable()) &&
        (excl(TARGET_NVD) || CreateCVETable(branches, TARGET_NVD)) &&
        (excl(TARGET_FSTEC) || CreateCVETable(branches, TARGET_FSTEC)) &&
        CreateBranchTables(branches, "packages");

    return success;
    }

//-----------------------------------------------------------------------------
// Delete the DB and create an empty one with the same name

bool TSpecificDBInterface::ClearDB(void){

    string err;
    bool success = true;

    printer.LineBegin("Cleaning " + database + " database");

    try{

        // Creating the DB connection
        if(!Connect(err)){
            throw runtime_error(err);
            }

        // Clearing the DB if exists
        if(DBExists(err)){

            if(!UseDB(err)){
                throw runtime_error(err);
                }

            vector<string> tables;
            if(!GetTableNames(tables, err)){
                throw runtime_error(err);
                }

            map<int, set<string>> hierarchy;
            const vector<set<string>> hierarchy_keys = {
                {"_issues_unfixed_", "_history_"},
                {"_issues", "_timelines"},
                {"_map", "_united", "_products"},
                {"_bin"},
                {"_src", "_import"}
                };

            for(const auto &table : tables){
                for(size_t i = 0; i < hierarchy_keys.size(); i++){
                    for(const auto &key : hierarchy_keys.at(i)){
                        if(table.find(key) != string::npos){
                            hierarchy[i].insert(table);
                            goto next_table;
                            }
                        }
                    }
                next_table:
                ;
                }

            for(size_t i = 1; i < hierarchy_keys.size(); i++){
                const auto it = hierarchy.find(i);
                if(it == hierarchy.end()){
                    continue;
                    }
                const auto &[_, tables_to_drop] = *it;
                for(const auto &table : tables_to_drop){
                    if(!DropTable(table, err)){
                        throw runtime_error(err);
                        }
                    }
                }
            }
        // DB does not exist - creating it
        else{
            if(!CreateDB(err) || !UseDB(err)){
                throw runtime_error(err);
                }
            }
        }
    catch(sql::SQLException &e){
        err = "Can't clean " + database + " database" + FormatSQLException(e);
        success = false;
        }
    catch(runtime_error &e){
        err = e.what();
        success = false;
        }

    if(!success){
        printer.Err(err);
        return false;
        }

    printer.Success();

    return true;
    }

//-----------------------------------------------------------------------------
// Call table-creation function for specified target

bool TSpecificDBInterface::CreateTargetTable(const vector<TBranch> &branches,
        const string &target, bool nested){

    if(!printer.IsSilent()){
        printer.Silent(nested);
        }

    bool success = false;

    if(target == TARGET_FSTEC || target == TARGET_NVD){
        success = CreateCVETable(branches, target);
        }
    else if(target == TARGET_CPE){
        success = CreateCPETable();
        }
    else if(target == "src" || target == "bin" || target == "packages"){
        success = CreateBranchTables(branches, target);
        }

    printer.ReSilent();

    return success;
    }

//-----------------------------------------------------------------------------
// Create table with given name and structure described in the query

bool TSpecificDBInterface::CreateTable(const string &table, const string &query,
        vector<string> child_tables){

    string err;

    printer.LineBegin("Creating " + table + " table");

    // Dropping child-tables first
    if(!child_tables.empty() && !DropChildTables(child_tables, err)){
        printer.Err(err);
        return false;
        }

    bool success = false;

    // Creating the table
    try{

        if(!Connect(err)){
            throw runtime_error(err);
            }

        stmt = con->createStatement();

        // Deleting table if already exists
        if(!DropTable(table, err)){
            throw runtime_error(err);
            }

        stmt->execute(query);
        success = true;
        }
    catch(sql::SQLException &e){
        err = FormatSQLException(e);
        }
    catch(runtime_error &e){
        err = e.what();
        }

    if(!success){
        err = "Can't create " + table + " table" +
            (err.empty() ? "" : "; " + err);
        printer.Err(err);
        return false;
        }

    printer.Success();

    // Child tables should be recreated if they existed before
    if(!child_tables.empty()){
        return RecreateChildTables(child_tables);
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Create specified tables

bool TSpecificDBInterface::RecreateChildTables(const vector<string> &tables){

    string branch_name;
    string target;

    for(const auto &table : tables){

        // Branch name and target are divided by "_" symbol
        const size_t delim_index = table.find("_");

        if((delim_index != string::npos) && ((delim_index + 1) < table.size())){

            branch_name = table.substr(0, delim_index);
            target = table.substr(delim_index + 1);

            if(branch_name == LOCAL_SYS && target == "timelines"){
                continue;
                }

            if(!CreateBranchTable(branch_name, target)){
                return false;
                }
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Create CVE table for the given target (FSTEC, NVD, ...)

bool TSpecificDBInterface::CreateCVETable(const vector<TBranch> &branches,
        const string &target){

    const string table = target + "_vul_import";

    string query = "CREATE TABLE " + table + "("
        "vul_year YEAR,"
        "vul_num VARCHAR(8),"; // Lengths of currently existing ids are <= 7,
                               // reserving 1 byte just in case
    if(target == TARGET_FSTEC){
        query += "vul_name TEXT,"
            "summary TEXT,"
            "software_list LONGTEXT,"
            "environment TEXT,"
            "cwe_id VARCHAR(16),"
            "published DATE,"
            "cvss_score_v2 FLOAT(3,1),"
            "cvss_score_v3 FLOAT(3,1),"
            "severity TEXT,"
            "solution TEXT,"
            "vul_status TEXT,"
            "exploit_status TEXT,"
            "fix_status TEXT,"
            "refs TEXT,"
            "patch TEXT,"
            "cve_ids TEXT,";
        }
    else if(target == TARGET_NVD){
        query += "software_list LONGTEXT,"
            "published DATE,"
            "modified DATE,"
            "cvss_score_v2 FLOAT(3,1),"
            "cvss_score_v3 FLOAT(3,1),"
            "cwe_id VARCHAR(16),"
            "refs TEXT,"
            "patch TEXT,"
            "summary TEXT,";
        }
    else{
        printer.Err("Unknown target table to create - \"" + table + "\"");
        return false;
        }

    query += "CONSTRAINT pk_" + table + " PRIMARY KEY (vul_year, vul_num))";

    vector<string> child_tables;

    // Vulnerabilities tables could not be dropped if there is an issues table
    if(target == TARGET_FSTEC || target == TARGET_NVD){
        string warn;
        for(const auto &branch : branches){
            const string child_table = branch.Name() + "_" + target + "_issues";
            if(!PushTable(child_tables, child_table, warn) && !warn.empty()){
                printer.Warn(warn);
                }
            }
        }

    return CreateTable(table, query, child_tables);
    }

//-----------------------------------------------------------------------------
// Create table for CPE dictionary

bool TSpecificDBInterface::CreateCPETable(void){

    const string table = "cpe_import";
    const string query = "CREATE TABLE " + table + "("
        "refs TEXT,"
        "part VARCHAR(1),"
        "vendor VARCHAR(128),"
        "product VARCHAR(128),"
        "non_free TINYINT UNSIGNED," // See CPE_SEP_NON_FREE
        "CONSTRAINT pk_" + table + " PRIMARY KEY (part, vendor, product)"
        ")";

    return CreateTable(table, query);
    }

//-----------------------------------------------------------------------------
// Create branch table for the given target (src or bin)

bool TSpecificDBInterface::CreateBranchTable(const string &branch_name,
        const string &target){

    const string table = branch_name + "_" + target;
    const vector<string> available_targets = {"src", "bin", "timelines"};

    // Target "packages" means "all tables for this branch"
    if(target == "packages"){
        for(auto &t : available_targets){
            if(branch_name == LOCAL_SYS && t == "timelines"){
                continue;
                }
            if(!CreateBranchTable(branch_name, t)){
                return false;
                }
            }
        return true;
        }
    else if(find(available_targets.begin(), available_targets.end(), target) ==
            available_targets.end()){
        printer.Err("Unknown target table to create - \"" + table + "\"");
        return false;
        }

    // Forming a query
    string query = "CREATE TABLE " + table + "(";
    if(target == "src"){
        query += "id INT UNSIGNED AUTO_INCREMENT,";
        }
    else if(target == "bin"){
        query += "src_id INT UNSIGNED NOT NULL,";
        }

    // Can't use src_id in timelines tables because these tables include info
    // about packages that can be either present or not present in src tables
    query += "name VARCHAR(128) COLLATE utf8mb4_0900_as_cs " +
        string(target != "timelines" ? "UNIQUE " : "") + "NOT NULL,";
    if(target == "src" || target == "timelines"){
        query += "ver VARCHAR(" + to_string(VER_LEN) + ") NOT NULL,";
        if(target == "src"){
            query += "rel VARCHAR(" + to_string(REL_LEN) + ") NOT NULL,"
                "nvd_name VARCHAR(128),"
                "fstec_name VARCHAR(128),"
                "url TEXT,"
                "cve_fixes TEXT,"
                "bdu_fixes TEXT,"
                "relatives TEXT,"
                "acl TEXT,"
                "CONSTRAINT pk_" + table + " PRIMARY KEY (id)";
            }
        else{
            query += "timestamp INT UNSIGNED NOT NULL,"
                "CONSTRAINT pk_" + table + " PRIMARY KEY (name, ver)";
            }
        }
    // target is "bin"
    else{
        query += "CONSTRAINT " + table + "_fk_" + branch_name + " "
            "FOREIGN KEY (src_id) REFERENCES "  + branch_name + "_src (id)";
        }
    query += ")";

    // Src table could not be dropped if there is a bin or issues tables
    vector<string> child_tables;
    if(target == "src"){
        const vector<string> suffixes = {"bin", "timelines", "nvd_issues",
            "fstec_issues"};
        for(const auto &suffix : suffixes){
            const string child_table = branch_name + "_" + suffix;
            string warn;
            if(!PushTable(child_tables, child_table, warn) && !warn.empty()){
                printer.Warn(warn);
                }
            }
        }

    return CreateTable(table, query, child_tables);
    }

//-----------------------------------------------------------------------------
// Create target table (src and bin) for each given branches

bool TSpecificDBInterface::CreateBranchTables(const vector<TBranch> &branches,
        const string &target){

    bool flag = false;
    bool success = true;
    const string master_branch = [&](void){
        for(const auto &branch : branches){
            if(branch.IsMasterBranch()){
                return branch.Name();
                }
            }
        return string();
        }();

    // For all branches
    for(const auto &branch : branches){
        if(branch.IsGhostBranch() &&
                (branch.Name() == master_branch || target != "timelines")){
            continue;
            }
        const string &branch_name =
            branch.IsLocalSys() && target != "timelines" ?
                LOCAL_SYS : branch.Name();
        string subtarget = "";
        if(target == "src" || target == "bin" || target == "timelines"){
            subtarget = target;
            }
        else if(target == branch_name || target == "packages"){
            subtarget = "packages";
            }
        else if(target == (branch_name + "_src") ||
                target == (branch_name + "_bin") ||
                target == (branch_name + "_timelines")){
            subtarget = target.substr(branch_name.size() + 1);
            }
        if(subtarget.size() > 0){
            flag = true;
            success = CreateBranchTable(branch_name, subtarget) && success;
            subtarget.clear();
            }
        }

    if(!flag){
        printer.Err("Unknown target" + (!target.empty() ? " " + target : ""));
        }

    return success;
    }

//-----------------------------------------------------------------------------
// Import list of package descriptions and their ACLs into the DB

bool TSpecificDBInterface::ImportPackages(TBranch &branch){

    string err;

    const array<string, 3> targets = {"src", "bin", "timelines"};
    const bool is_local_sys = branch.IsLocalSys();
    const auto &src = branch.Src();
    const auto &bin = branch.Bin();
    const auto &timelines = branch.TimeLines();
    const auto &acl = branch.ACL();
    const string src_table = (is_local_sys ? LOCAL_SYS : branch.Name()) + "_src";

    for(const auto &target : targets){

        if(branch.IsGhostBranch() && target != "timelines"){
            continue;
            }

        bool success = false;
        bool critical_err = false;
        const string table = (is_local_sys && target != "timelines" ?
            LOCAL_SYS : branch.Name()) + "_" + target;
        const string fields = [&](void){
            string s;
            if(target == "src"){
                s = "(name, ver, rel, ";
                for(auto &m : KeysV(vul_markers, true)){
                    s += (m + "_fixes, ");
                    }
                s += "url, relatives, acl) ";
                }
            else if(target == "bin"){
                s = "(name, src_id) ";
                }
            else{
                s = "(ver, timestamp, name) ";
                }
            return s;
            }();

        printer.LineBegin("Importing data into " + table + " table");

        try{

            if(!Connect(err)){
                throw runtime_error(err);
                }

            stmt = con->createStatement();

            // Cleaning the table or creating a new empty one
            if(TableExists(table, err)){
                if(!appending){
                    stmt->execute("DELETE FROM " + table);
                    }
                }
            else{
                printer.Numb(true);
                const bool created = CreateBranchTable(branch.Name(), target);
                printer.Numb(false);
                if(!created){
                    throw runtime_error(err);
                    }
                }

            // Forming a query for adding descriptions

            string query = "INSERT INTO " + table + " " + fields + "VALUES ";
            string values;

            if(target == "src"){

                for(const auto &[src_name, src_desc] : src){

                    if(src_name.empty() || src_desc.size() < PACK_DESC_SIZE){
                        continue;
                        }

                    const auto &src_ver = src_desc.at(PACK_VER);
                    const auto &src_rel = src_desc.at(PACK_REL);
                    const auto &src_fixed_cve = src_desc.at(PACK_FIXED_CVE);
                    const auto &src_fixed_bdu = src_desc.at(PACK_FIXED_BDU);
                    const auto &src_url = src_desc.at(PACK_URL);
                    const auto &src_relatives = src_desc.at(PACK_RELATIVES);

                    const string ver_len = to_string(VER_LEN);

                    values += "("
                        "'" + src_name + "'," +
                        "SUBSTRING('" + src_ver + "', 1, " + ver_len + "),"
                        "SUBSTRING('" + src_rel + "', 1, " + ver_len + "),"
                        "'" + src_fixed_cve + "',"
                        "'" + src_fixed_bdu + "',"
                        "'" + src_url       + "',"
                        "'" + src_relatives + "',";

                    string package_acl;
                    const auto acl_it = acl.find(src_name);
                    if(acl_it != acl.end()){
                        const auto &acl_entries = acl_it->second;
                        for(const auto &user : acl_entries){
                            if(!package_acl.empty()){
                                package_acl += ",";
                                }
                            package_acl += user;
                            }
                        }

                    values += "'" + package_acl + "'),";
                    }
                }

            else if(target == "bin"){
                for(const auto &[src_name, bin_names] : bin){
                    for(const auto &bin_name : bin_names){
                        values += "('" + bin_name + "', "
                            "(SELECT id FROM " + src_table + " "
                                "WHERE name = '" + src_name + "')),";
                        }
                    }
                }

            // target is "timelines"
            else{
                for(const auto &[src_name, versions] : timelines){
                    if(src_name.empty()){
                        continue;
                        }
                    for(const auto &[ver, timestamp] : versions){
                        if(ver.empty()){
                            continue;
                            }
                        values += "(SUBSTRING('" + ver + "', "
                            "1, " + to_string(VER_LEN) + "),"
                            "'" + to_string(timestamp) + "',"
                            "'" + src_name + "'),";
                        }
                    }
                }

            if(values.empty()){
                throw runtime_error("Empty " + target + " list");
                }

            values.pop_back();
            query += values;

            // Sending formed query into the DB
            stmt->execute(query);
            success = true;
            }
        catch(sql::SQLException &e){
            err = (!err.empty() ? err + "; " : "") + FormatSQLException(e);
            critical_err = true;
            }
        catch(runtime_error &e){
            err  = e.what();
            err += !err.empty() ? "; " : "";
            err += "Can't create \"" + table + "\" table";
            }

        if(success){
            printer.Success();
            }
        else if(!critical_err && target == "timelines"){
            printer.Warn(err);
            }
        else{
            printer.Err(err);
            return false;
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Form CSV import query

string TSpecificDBInterface::FormCSVImportQuery(const string &csv_path,
        const string &table) const{

    const string query = "LOAD DATA LOCAL INFILE '" + csv_path + "' "
        "INTO TABLE " + table + " "
        "FIELDS TERMINATED BY ',' "
        "ENCLOSED BY '\"' "
        "LINES TERMINATED BY '\n' ";

    return query;
    }

//-----------------------------------------------------------------------------
// Import data in CSV format into the DB

bool TSpecificDBInterface::ImportCSV(const vector<TBranch> &branches,
        const string &target, const string &csv_path, bool force_clean){

    string err;

    // Checking available targets
    string msg = "Importing ";
    if(target == TARGET_FSTEC){
        msg += "FSTEC vulnerabilities list ";
        }
    else if(target == TARGET_NVD){
        msg += "NVD CVE list ";
        }
    else if(target == TARGET_CPE){
        msg += "CPE dictionary ";
        }
    else{
        return false;
        }
    msg += "to " + Name();

    printer.LineBegin(msg);

    // CSV file path needed
    if(csv_path.empty()){
        printer.Err("Path to CSV file not specified");
        return false;
        }

    try{

        if(!Connect(err)){
            printer.Err(err);
            return false;
            }

        const string table = target + (target != TARGET_CPE ? "_vul" : "") +
            "_import";

        // Cleaning the table or creating new empty table
        if(TableExists(table, err)){
            // NVD CVE lists comes in many files, so nvd_vul_import table
            //  should not be cleaned before each import
            if(target != TARGET_NVD || force_clean){
                if(!appending){
                    stmt->execute("DELETE FROM " + table);
                    }
                }
            }
        else{
            CreateTargetTable(branches, target, true);
            }

        // Inserting data
        const string query = FormCSVImportQuery(csv_path, table);
        stmt = con->createStatement();
        stmt->execute(query);
        }
    catch(sql::SQLException &e){
        err = "Can't import CSV to " + Name() + "; " + FormatSQLException(e);
        printer.Err(err);
        return false;
        }

    printer.Success();

    return true;
    }

//-----------------------------------------------------------------------------
