/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2022 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    <thread>
#include    <chrono>

#include    "importer.h"
#include    "tree.h"

//-----------------------------------------------------------------------------
// Check if source with given name is excluded

bool TImporter::ExcludedVulSrc(const string &name) const{

    const auto &excl = conf.ExclVulSrc();

    return find(excl.begin(), excl.end(), name) != excl.end();
    }

//-----------------------------------------------------------------------------
// Configure the program

bool TImporter::Configure(void){

    string err;

    // Reading the conf file
    if(!conf.Configure(err)){
        printer.Err(err);
        return false;
        }

    // Setting up the DB connection params
    if(!db.Configure(conf.DBConParams(), err)){
        printer.Err(err);
        return false;
        }

    bool got_local_sys = false;
    bool got_sec_for_master_branch = false;
    const auto &master_branch = conf.CommonParams().at("master_branch");

    // Converting branch params to TBranch type
    for(const auto &[branch_name, branch_params] : conf.Branches()){
        TBranch branch;
        // Checking for the "local_sys" mark
        if(branch_name == LOCAL_SYS){
            if(got_local_sys){
                printer.Err("More than one branch is marked as a local system");
                return false;
                }
            const auto branch_param = branch_params.find("branch");
            if(branch_param != branch_params.end()){
                const auto &[_, b] = *branch_param;
                branch.WriteName(b.empty() ? master_branch : b);
                }
            else{
                branch.WriteName(master_branch);
                }
            branch.MakeLocal();
            got_local_sys = true;
            }
        else{
            branch.WriteName(branch_name);
            }
        // Handling the rest of the parameters
        branch.WriteBaseDir(branch_params.at("base_dir"));
        branch.WriteSRPMSDir(branch_params.at("srpms_dir"));
        branch.WriteRPMSDirs(branch_params.at("rpms_dirs"));
        branch.WriteBinList(branch_params.at("bin_list"));
        if(!got_sec_for_master_branch && branch.Name() == master_branch){
            branch.MakeMaster();
            got_sec_for_master_branch = true;
            }
        branches.push_back(branch);
        }

    if(branches.empty()){
        printer.Err(
            "There is not a single section with a description of a branch");
        return false;
        }

    // Create a ghost branch if there is no section for a master branch
    if(all_of(branches.begin(), branches.end(),
            [](TBranch b){return !b.IsMasterBranch();})){
        TBranch branch;
        branch.WriteName(master_branch);
        branch.MakeGhost();
        branches.push_back(branch);
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Recreate the DB and import everything that could be imported

bool TImporter::PrepareDB(void){

    const bool success =
        RecreateDBStruct() &&
        ImportPackages() &&
        ImportXmlOrJson(TARGET_CPE) &&
        ImportXmlOrJson(TARGET_NVD) &&
        ImportXmlOrJson(TARGET_FSTEC);

    if(!success){
        printer.Err("Database was not prepared");
        }

    return success;
    }

//-----------------------------------------------------------------------------
// Get names of all available branches

set<string> TImporter::NamesOfAllBranches(void) const{

    set<string> names;

    for(const auto &branch : branches){
        names.insert(branch.Name());
        }

    return names;
    }

//-----------------------------------------------------------------------------
// Import lists of packagese for all considered branches

bool TImporter::ImportPackages(const set<string> &selected_branches){

    #define SLEEP_SEC 5

    // Checking selected branches
    if(!selected_branches.empty()){
        const auto all_branches = NamesOfAllBranches();
        for(const auto &selected_branch : selected_branches){
            if(all_branches.find(selected_branch) != all_branches.end()){
                continue;
                }
            printer.Err("There is no '" + selected_branch + "' branch available");
            return false;
            }
        }

    const array<string, 3> targets = {"bin", "acl", "timelines"};

    // For all branches stated in the configuration file
    for(auto &branch : branches){

        // Skipping branches that were not selected (if there is a selection)
        if(!selected_branches.empty() &&
                selected_branches.find(branch.Name()) == selected_branches.end()){
            continue;
            }

        set<string> missing_packages0;

        // While there are some missing packages
        while(true){

            // Data for already present packages will not be rewritten
            const auto excluded_src = Keys(branch.Src());

            // Loading every target list to the memory
            for(const auto &target : targets){
                const string &base_dir = (target != "bin") ?
                    conf.CommonParams().at("download") :
                    conf.CommonParams().at("basealt");
                if(!branch.Import(target, base_dir, excluded_src)){
                    return false;
                    }
                }

            const string &base_dir = conf.CommonParams().at("basealt");

            // Retrieving metadata of rpm packages
            const auto [missing_packages, ok] =
                    branch.ImportPackagesMetadata(base_dir, excluded_src);
            if(!ok){
                return false;
                }
            if(missing_packages.empty() || run_mode.Flag(DEBUG_FLAG)){
                break;
                }
            // Stop trying if a set of missing packages is the same
            if(missing_packages == missing_packages0){
                return false;
                }

            // Erasing data related to missing packages
            branch.Erase(missing_packages);

            missing_packages0 = missing_packages;

            // Waiting for a few seconds before a next take
            printer.LineEnd("~~~ Sleeping for a " + to_string(SLEEP_SEC) + " "
                "second(s) before a next try ~~~");
            this_thread::sleep_for(chrono::seconds(SLEEP_SEC));
            }

        // Finding relations between packages
        if(!branch.DetectRelatives()){
            return false;
            }

        // Importing package descriptions to the database
        if(!db.ImportPackages(branch)){
            return false;
            }

        // Releasing the memory
        branch.Clear();
        }

    #undef SLEEP_SEC

    return true;
    }

//-----------------------------------------------------------------------------
// Remove non-printable characters from the specified text file and replace all
// unicode-spaces contained in it with the ASCII space character

bool TImporter::HandleDisallowedSymbols(const string &file_full_path){

    printer.LineBegin("Removing/replacing disallowed symbols from "
        "\"" + file_full_path + "\" file");

    const string unicode_spaces =
        "\\U00A0\\U1680\\U180E\\U2000\\U2001\\U2002\\U2003\\U2004\\U2005\\U2006"
        "\\U2007\\U2008\\U2009\\U200A\\U200B\\U202F\\U205F\\U3000\\UFEFF";
    const string replace_unicode_spaces =
        "s/['$(printf \"%b\" \"" + unicode_spaces + "\")']/ /g";
    const string remove_non_printable_symbols = "s/[^[:print:]]//g";
    const string cmd = "sed -i.bak '" + remove_non_printable_symbols + "; " +
        replace_unicode_spaces + "' " + file_full_path;
    const string suppress_err = "2>/dev/null";

    const auto ret_code = system((cmd + " " + suppress_err).c_str());

    if(WEXITSTATUS(ret_code) != 0){
        printer.Err("Command \"" + cmd + "\", "
            "error code " + to_string(ret_code));
        return false;
        }

    printer.Success();
    return true;
    }

//-----------------------------------------------------------------------------
// Import data from XML or JSON file

bool TImporter::ImportXmlOrJson(const string &target, const string &file_path,
        const bool force_clean){

    // Checking the target
    if(target != TARGET_NVD && target != TARGET_FSTEC && target != TARGET_CPE){
        printer.Err("Wrong import target \"" + target + "\"");
        return false;
        }

    const string data_src = (target == TARGET_CPE ? TARGET_NVD : target);

    // Skipping the operation if given target is excluded
    if(ExcludedVulSrc(data_src)){
        const string msg = "Source '" + data_src + "' is excluded";
        if(displayed_notes.find(msg) == displayed_notes.end()){
            displayed_notes.insert(msg);
            printer.LineEnd("[NOTE: " + msg + "]");
            }
        return true;
        }

    // Filling the maps of excluded products if not filled yet
    const string excl_list = (target == TARGET_FSTEC ? FSTEC_EXCL : CPE_EXCL);
    const auto it = excluded_products.find(excl_list);
    if(it == excluded_products.end() && !GetExcludedProducts(excl_list)){
        return false;
        }

    string file_full_path;

    // Using path from Conf-file if path not given
    if(file_path.empty()){
        const string &download_path = conf.CommonParams().at("download");
        if(target == TARGET_FSTEC){
            file_full_path = download_path + FSTEC_VUL;
            }
        else if(target == TARGET_NVD){
            const int current_year = CurrentYear();
            bool first_insert = true;
            // Calling oneself with every default filename
            for(int i = 2002; i <= (current_year + 1); i++){
                file_full_path = download_path + "nvdcve-1.1-" +
                    (i > current_year ? "recent" : to_string(i)) +
                    "." + import_format[target];
                if(!CheckFile(file_full_path)){
                    printer.Warn("Can't find \"" + file_full_path + "\" file");
                    continue;
                    }
                if(!ImportXmlOrJson(target, file_full_path, first_insert)){
                    return false;
                    }
                first_insert = false;
                }
            return true;
            }
        else{
            file_full_path = download_path + CPE_DICT;
            }
        }
    else{
        file_full_path = file_path;
        }

    // The FSTEC vulnerability list can contain non-printable characters that
    // the libxml++2 doesn't allow
    if(target == TARGET_FSTEC && !HandleDisallowedSymbols(file_full_path)){
        return false;
        }

    printer.LineBegin("Extracting subtree "
        "from " + StripPath(file_full_path) + " tree");

    // Validating the file path
    if(!CheckFile(file_full_path)){
        const string msg = "There is no \"" + file_full_path + "\" file "
            "to import";
        if(target == TARGET_FSTEC){
            printer.Warn(msg);
            }
        else{
            printer.Err(msg);
            }
        return false;
        }

    auto err_msg = [](string err){
        replace(err.begin(), err.end(), '\n', ' ');
        return err;
        };

    tree::TTree *ptree = NULL;
    string csv_file_path;
    bool abort = false;

    // Parsing the document
    try{

        string root;
        vector<tree::TBranchDraft> branch_drafts;
        string err;

        // Constructing a tree of needed nodes
        if(import_format[target] == "xml"){
            ptree = new tree::TXmlTree();
            tie(root, branch_drafts) = FormXmlBranchDrafts(target, err);
            }
        else{
            ptree = new tree::TJsonTree();
            tie(root, branch_drafts) = FormJsonBranchDrafts(target, err);
            }
        if(!ptree){
            throw string("Can't spawn new TTree object");
            }
        if(root.empty() || branch_drafts.empty()){
            throw err;
            }
        if(!ptree->Construct(root, branch_drafts)){
            throw "Can't create tree of needed nodes "
                "of \"" + file_full_path + "\" " + import_format[target];
            }
        // Extracting the data using constructed tree
        if(!ptree->Select(file_full_path, err)){
            throw err;
            }

        // If there is a data to import
        if(!ptree->SelectedData().empty()){
            printer.Success();
            // Making some kind of adjustments of the selected data
            if(!Preprocess(target, ptree->SelectedData(), excluded_products, err)){
                throw err;
                }
            // Producing CSV string out of selected data
            csv_file_path = GenFileName(file_full_path);
            if(!WriteToFile(
                    csv_file_path,
                    ptree->SelectedData().Stringify(SEL_L_BORDER, SEL_R_BORDER, SEL_DELIM)
                    )){
                csv_file_path.clear();
                }
            }
        else{
            throw string("Empty data selection");
            }

        if(ptree){
            delete ptree;
            }
        }
    catch(const string &e){
        printer.Err(err_msg(e));
        abort = true;
        }
    catch(const exception &e){
        printer.Err(err_msg(e.what()));
        abort = true;
        }
    catch(...){
        printer.Err();
        abort = true;
        }

    if(abort){
        if(ptree){
            delete ptree;
            }
        return false;
        }

    if(csv_file_path.empty()){
        printer.Err("Can't create CSV file for subsequent import operation");
        return false;
        }

    // Importing tmp CSV file (that contains selected data) into the DB
    const bool import_status = db.ImportCSV(branches, target, csv_file_path,
        force_clean);

    // Removing tmp CSV file
    const int remove_status = remove(csv_file_path.c_str());
    if(remove_status != 0){
        printer.Warn("Can't remove temp file");
        }

    return import_status;
    }

//-----------------------------------------------------------------------------
// Import data from many XML or JSON files

bool TImporter::ImportXmlOrJson(const string &target,
        const set<string> &file_paths, const bool force_clean){

    if(file_paths.empty()){
        return ImportXmlOrJson(target, "", force_clean);
        }

    for(const auto &file_path : file_paths){
        if(!ImportXmlOrJson(target, file_path, force_clean)){
            return false;
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Fill a map of excluded products with values from a specified CSV file, where
// each line is "<vendor>, <product_0> <product_1> .. <product_n>" and each
// product name can be marked as excluded one by the "!" front character

bool TImporter::GetExcludedProducts(const string &path){

    printer.LineBegin("Getting excluded products from \"" + path + "\" file");

    // After calling the function "GetExcludedProducts", there must be a value
    // for the key passed by the "path" argument
    excluded_products[path];

    if(!CheckFile(path)){
        printer.Warn("File doesn't exist");
        return true;
        }

    TFileInMem file;

    if(!file.ReadFileToMem(path)){
        printer.Err("Can't read file");
        return false;
        }

    string line;
    int line_n = 0;
    bool header = true;
    istringstream iss(file.data);

    // For all lines of a file that contains excluded products
    while(getline(iss, line)){

        line_n++;
        Trim(line);
        if(line.empty()){
            continue;
            }

        if(header){
            header = false;
            continue;
            }

        const size_t comma_pos = line.find(',');
        if((comma_pos == 0) || (comma_pos == line.size() - 1) ||
                (comma_pos == string::npos) || (line.rfind(',') != comma_pos)){
            printer.Err(FormatViolation(path, line_n));
            return false;
            }

        string vendor = line.substr(0, comma_pos);
        Trim(vendor);
        string product_names = line.substr(comma_pos + 1);
        Trim(product_names);

        for(const auto &product_name : Split(product_names)){
            // If product is exceptional
            if(product_name.front() == '!'){
                const auto product_name_clean = product_name.substr(1);
                if(product_name_clean.empty() || product_name_clean == "*" ||
                        vendor == "*"){
                    printer.Err(FormatViolation(path, line_n));
                    return false;
                    }
                auto &[excl_product_names, excp_product_names] =
                    excluded_products[path][vendor];
                excl_product_names.insert("*");
                excp_product_names.insert(product_name_clean);
                }
            // Product is excluded
            else{
                if(vendor == "*" && product_name == "*"){
                    printer.Err(FormatViolation(path, line_n));
                    return false;
                    }
                excluded_products[path][vendor].first.insert(product_name);
                }
            }
        }

    printer.Success();

    return true;
    }

//-----------------------------------------------------------------------------
