/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2022 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    <sstream>
#include    <filesystem>
#include    <algorithm>
#include    <regex>

#include    "lists.h"
#include    "related_package_names.h"
#include    "rpm.h"

namespace fs = filesystem;

//-----------------------------------------------------------------------------
// Assign a vector of elements of a split of a given str to the rpms_dirs
// vector

void TBranch::WriteRPMSDirs(const string &s){

    rpms_dirs.clear();

    for(const auto &path : Split(s, "//")){
        if(path.empty()){
            continue;
            }
        // There are no relative paths allowed
        rpms_dirs.push_back(path.front() == '/' ? path : ("/" + path));
        }
    }

//-----------------------------------------------------------------------------
// Import a bin list or an acl or timelines of this branch

bool TBranch::Import(const string &target, const string &base_dir,
        const set<string> &excluded_src){

    if(is_ghost_branch && target == "bin"){
        return true;
        }

    const string msg = is_local_sys && target == "bin" ?
        "Forming a list of installed packages" :
        "Parsing " + target + (target == "bin" ? " list" : "") +
            " of " + name + " branch";

    printer.LineBegin(msg);

    bool success = false;
    vector<string> import_file_paths;
    string prefix;
    array<string, 3> errors;
    function<bool(const TFileInMem&, const set<string>&, string&)> import_func;
    TFileInMem content_of_import_file;

    // Forming a full path of an import file
    if(target == "bin"){
        import_file_paths.push_back(BinListPath(base_dir, errors.at(0)));
        import_func = [=](const TFileInMem &f, const set<string> &excl, string &e){
            return this->ImportBinList(f, excl, e);
            };
        }
    else if(target == "acl"){
        import_file_paths.push_back(ACLPath(base_dir, errors.at(0)));
        import_func = [=](const TFileInMem &f, const set<string> &excl, string &e){
            return this->ImportACL(f, excl, e);
            };
        }
    else if(target == "timelines"){
        import_file_paths = TimeLinesPaths(base_dir, errors.at(0));
        import_func = [=](const TFileInMem &f, const set<string> &excl, string &e){
            return this->ImportTimeLine(f, excl, e);
            };
        // Last element is a common part of all the paths
        if(import_file_paths.size() > 1){
            prefix = import_file_paths.back();
            import_file_paths.pop_back();
            }
        }
    else{
        errors[0] = "Wrong import target \"" + target + "\"";
        goto ImportReturn;
        }

    if(!is_local_sys &&
            (import_file_paths.empty() || import_file_paths.front().empty())){
        errors[0] = "Can't determine " + target + " file path(s) for the " +
            name + " branch";
        goto ImportReturn;
        }

    // Handling import files one by one
    for(const auto &import_file_path : import_file_paths){
        // Reading the import file to memory beforehand
        if(!is_local_sys || target != "bin"){
            if(import_file_path.empty()){
                errors[1] = "There is no " + target + " file path";
                break;
                }
            success = content_of_import_file.ReadFileToMem(prefix.empty() ?
                import_file_path : (prefix + import_file_path));
            if(!success){
                errors[1] = "Can't read \"" + import_file_path + "\" file";
                break;
                }
            }
        // Converting the data to the internal representation
        success = import_func(content_of_import_file, excluded_src, errors[2]);
        if(!success){
            break;
            }
        }

    ImportReturn:

    // Forming an overall error message
    string err_msg;
    for(const auto &err : errors){
        if(!err.empty()){
            err_msg += (err_msg.empty() ? "" : "; ") + err;
            }
        }

    if(!success){
        if(target == "acl" || (target == "timelines" && !IsMasterBranch())){
            if(err_msg.empty()){
                err_msg = "Can't import " + target + ", this is acceptable for "
                    "non master branch" ;
                }
            printer.Warn(err_msg);
            return true;
            }
        printer.Err(err_msg);
        return false;
        }

    printer.Success();

    return true;
    }

//-----------------------------------------------------------------------------
// Get IDs of fixed vulnerabilities and URLs from src or bin packages, return
// a set of src names of missing packages and a flag that is false in case of
// a critical error

pair<set<string>, bool> TBranch::ImportPackagesMetadata(
        const string &alt_mnt_point, const set<string> &excluded_src){

    if(is_ghost_branch){
        return {{}, true};
        }

    printer.LineBegin("Getting metadata from " +
        (is_local_sys ?
            "installed packages of the local system" :
            "src/bin RMP packages of the " + name + " branch"));

    string err;

    const auto srpms_dir_to_process = is_local_sys ? string() :
        SRPMSDirPath(alt_mnt_point, err);
    const auto rpms_dirs_to_process = is_local_sys ? vector<string>() :
        RPMSDirPaths(alt_mnt_point, err);

    if(!err.empty()){
        printer.Err(err);
        return {{}, false};
        }

    vector<string> warn;

    const auto [success_count, missing_packages] =
        GetMetadata(
            srpms_dir_to_process,
            rpms_dirs_to_process,
            src,
            bin,
            is_local_sys,
            excluded_src,
            warn);

    const auto packages = Keys(src);
    set<string> not_excluded_packages;
    set_difference(
        packages.begin(),
        packages.end(),
        excluded_src.begin(),
        excluded_src.end(),
        inserter(not_excluded_packages, not_excluded_packages.end()));

    const int n_processed_packages =
        not_excluded_packages.size() - missing_packages.size();

    if(n_processed_packages == 0){
        printer.LineAddErr(warn);
        printer.Err("Zero processed packages");
        return {missing_packages, false};
        }

    printer.LineAddWarn(warn);

    string msg = to_string(n_processed_packages) + "/" +
        to_string(not_excluded_packages.size()) + " packages inspected";
    for(auto &[k, v] : success_count){
        msg += ", got " + k + " from " + to_string(v) + " packages";
        }
    if(!missing_packages.empty()){
        printer.Warn(msg);
        }
    else{
        printer.Success(msg);
        }

    return {missing_packages, true};
    }

//-----------------------------------------------------------------------------
// Form a base path of this branch

string TBranch::BasePath(const string &alt_mnt_point){

    return base_dir.empty() ?
        DirEnding(alt_mnt_point) + "ALT/" + name + "/" :
        DirEnding(base_dir);
    }

//-----------------------------------------------------------------------------
// Form full path for a given subpath inside a base directory

string TBranch::DirInsideBase(const string &alt_mnt_point,
        const string &subpath){

    // Need a name of the branch and a basealt/custom mount point
    return (name.empty() || (alt_mnt_point.empty() && base_dir.empty())) ?
        "" :
        BasePath(alt_mnt_point) + subpath;
    }

//-----------------------------------------------------------------------------
// Form a full file path of a list of packages of this branch
// (/space/ALT/<branch name>/files/list/bin.list)

string TBranch::BinListPath(const string &alt_mnt_point, string &err){

    // If custom bin_list path is given or otherwise
    string path = !bin_list.empty() ?
        bin_list :
        DirInsideBase(alt_mnt_point, "files/list/bin.list");

    if(path.empty()){
        err = "Can't form bin_list path";
        return "";
        }

    path = FormatFilePath(path, NOT_A_DIR);

    // Checking the read permission
    if(!CheckFile(path)){
        err = "Can't read \"" + path + "\"";
        path.clear();
        }

    return path;
    }

//-----------------------------------------------------------------------------
// Form full path of a directory that contain src packages

string TBranch::SRPMSDirPath(const string &alt_mnt_point, string &err){

    // If custom srpms_dir path is given or otherwise
    string path = (srpms_dir.empty() && rpms_dirs.empty()) ?
        DirInsideBase(alt_mnt_point, "files/SRPMS/") :
        srpms_dir;

    if(path.empty()){
        return "";
        }

    path = FormatFilePath(path, IS_DIR);

    // Checking the read permission
    if(!CheckFile(path)){
        err = "Can't read \"" + path + "\"";
        path.clear();
        }

    return path;
    }

//-----------------------------------------------------------------------------
// Format paths of directories that contain bin packages

vector<string> TBranch::RPMSDirPaths(const string &alt_mnt_point, string &err){

    vector<string> paths;

    for(const auto &path0 : rpms_dirs){

        if(path0.empty()){
            continue;
            }

        const string path = FormatFilePath(path0, IS_DIR);

        // Checking the read permission
        if(!CheckFile(path)){
            err = "Can't read \"" + path + "\"";
            continue;
            }

        paths.push_back(path);
        }

    return paths;
    }

//-----------------------------------------------------------------------------
// Form a full file path of an acl file of this branch

string TBranch::ACLPath(const string &base_dir, string &err){

    string branch_name = BaseDir().empty() ? Name() : StripPath(BaseDir());

    if(branch_name.back() == '/'){
        branch_name.pop_back();
        }

    const string acl_name = "list.packages." + ToLower(branch_name);
    const string acl_path = FormatFilePath(base_dir, IS_DIR) +
        "acl/" + acl_name;

    if(!CheckFile(acl_path)){
        err = "Can't read " + acl_path;
        return "";
        }

    return acl_path;
    }

//-----------------------------------------------------------------------------
// Form a vector of full file paths of timeline files of this branch

vector<string> TBranch::TimeLinesPaths(const string &base_dir, string &err){

    vector<string> timelines_paths;
    string branch_name = BaseDir().empty() ? Name() : StripPath(BaseDir());

    if(branch_name.back() == '/'){
        branch_name.pop_back();
        }

    const string timelines_base_dir = FormatFilePath(base_dir, IS_DIR) +
        "timelines/" + ToLower(branch_name) + "/index/src/";

    if(!fs::exists(timelines_base_dir) || !fs::is_directory(timelines_base_dir)){
        return {};
        }

    const string wanted = "d-t-s-evr.list";

    // Walking through dirs of timelines
    for(auto &p : fs::recursive_directory_iterator(timelines_base_dir)){
        if(!is_regular_file(p) || p.path().filename().string() != wanted){
            continue;
            }
        const string &path = p.path().string();
        if(!CheckFile(path)){
            err += (err.empty() ? "Can't read ": ", ") + path;
            continue;
            }
        // Saving a file path minus base dir path in the output vector
        timelines_paths.push_back(path.substr(timelines_base_dir.size()));
        }

    // Saving a base dir path as a last element of the output vector
    if(!timelines_paths.empty()){
        timelines_paths.push_back(timelines_base_dir);
        }

    return timelines_paths;
    }

//-----------------------------------------------------------------------------
// Determine a src name and a ver-rel from a given src.rpm name

const set<string> extensions = {".src.rpm", ".src"};

bool DecomposeRPM(const string &src_rpm_name, string &src_name, string &ver,
        string &rel){

    size_t ext_index = string::npos;

    // Getting an index of the package extension
    for(const auto &ext : extensions){
        ext_index = src_rpm_name.rfind(ext);
        if(ext_index != string::npos &&
                ext_index == (src_rpm_name.size() - ext.size())){
            break;
            }
        }
    if(ext_index == string::npos){
        ext_index = src_rpm_name.size();
        }

    // 1st iteration - getting index of a release part
    // 2nd iteration - getting index of a version part
    array<size_t, 2> indexes = {ext_index, string::npos};
    for(int i = 0; i < 2; i++){
        indexes[i] = src_rpm_name.rfind('-', indexes[0] - 1);
        if(indexes[i] == string::npos){
            return false;
            }
        }

    // Getting a version and a release parts
    const auto &rel_index = indexes[0];
    const auto &ver_index = indexes[1];
    const size_t ver_len = rel_index - ver_index - 1;
    const size_t rel_len = ext_index - rel_index - 1;
    ver = src_rpm_name.substr(ver_index + 1, ver_len);
    rel = src_rpm_name.substr(rel_index + 1, rel_len);

    // Getting a src package name
    src_name = src_rpm_name.substr(0, ver_index);

    return true;
    }

//-----------------------------------------------------------------------------
// Get a number (starting from 0) of a last column of a given line of words
// separated with spaces; That column contains a src name (1 - if given line
// is a pair of a bin name and a src name, 4 - if given line is a line of an
// ALT Linux bin.list)

const set<int> src_name_valid_pos = {1, 4};

size_t SrcNamePos(const string &line){

    size_t counter = string::npos;
    size_t index = string::npos;
    string word;

    for(;;){
        index = GetWordFromLine(line, word, index + 1);
        if(index == string::npos){
            break;
            }
        counter++;
        }

    if(src_name_valid_pos.find(counter) == src_name_valid_pos.end()){
        return string::npos;
        }

    return counter;
    }

//-----------------------------------------------------------------------------
// Check unwanted suffixes

const set<string> unwanted_suffixes = {"-common", "-debuginfo", "-devel",
    "-doc", "-data"};

bool GotUnwantedSuffix(const string &package_name){

    for(const auto &suffix : unwanted_suffixes){
        const int delta = package_name.size() - suffix.size();
        if(delta > 0 &&
                package_name.compare(delta, suffix.size(), suffix) == 0){
            return true;
            }
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Add formed package description

void TBranch::AddPackageDesc(const string &src_name, const string &bin_name,
        const string &ver, const string &rel, const bool skipped_desc){

    const map<int, string> src_desc = {
        {PACK_VER, ver},
        {PACK_REL, rel},
        {PACK_FIXED_CVE, ""},
        {PACK_FIXED_BDU, ""},
        {PACK_URL, ""},
        {PACK_RELATIVES, ""}
        };

    auto &src_map = skipped_desc ? src_skipped : src;
    auto &bin_map = skipped_desc ? bin_skipped : bin;

    src_map[src_name] = src_desc;
    bin_map[src_name].insert(bin_name);
    }

//-----------------------------------------------------------------------------
// Restore records for those src packages whose bin packages all have unwanted
// suffixes

void TBranch::RestoreFromSkipped(void){

    for(const auto &[src_name, desc] : src_skipped){
        const auto it = src.find(src_name);
        if(it == src.end()){
            const auto it = bin_skipped.find(src_name);
            if(it == bin.end()){
                continue;
                }
            const auto &[_, bin_names] = *it;
            src[src_name] = desc;
            bin[src_name] = bin_names;
            }
        }

    src_skipped.clear();
    bin_skipped.clear();
    }

//-----------------------------------------------------------------------------
// Form a list of local system packages using RPM

bool TBranch::LocalSysBinList(const set<string> &excluded_src, string &err){

    const auto &[packages, err_msg] = GetSysPackages();

    if(!err_msg.empty()){
        err = err_msg;
        return false;
        }

    auto wrong_format_msg = [](const string &package_name){
        return "Wrong format of the package name \"" + package_name + "\"";
        };

    // For every src package and ints binary packages...
    for(const auto &[src_package, bin_packages] : packages){

        string src_name, src_ver, src_rel;

        if(!DecomposeRPM(src_package, src_name, src_ver, src_rel)){
            err = wrong_format_msg(src_package);
            return false;
            }

        // For every bin package of this src package...
        for(const auto &bin_package : bin_packages){

            string bin_name, bin_ver, bin_rel;

            if(!DecomposeRPM(bin_package, bin_name, bin_ver, bin_rel)){
                err = wrong_format_msg(bin_name);
                return false;
                }

            // Skipping some of the binary packages
            const bool skipped_desc = GotUnwantedSuffix(bin_name);

            AddPackageDesc(src_name, bin_name, src_ver, src_rel, skipped_desc);
            }
        }

    RestoreFromSkipped();

    return true;
    }

//-----------------------------------------------------------------------------
// Import descriptions of bin packages of this branch

#define BIN_NAME_COL 0 // Number of a column in bin list that holds a name
                       // of a bin package

bool TBranch::ImportBinList(const TFileInMem &file,
        const set<string> &excluded_src, string &err){

    // Forming a list of local system packages using RPM, not by parsing a text
    // file prepared in advance
    if(is_local_sys){
        return LocalSysBinList(excluded_src, err);
        }

    // Column number in bin-list that holds a src name (will be redefined later)
    size_t SRC_NAME_COL = string::npos;

    unsigned line_n = 0;
    string line, word;
    istringstream iss(file.data);

    // Parsing lines of an import file one by one
    while(getline(iss,line)){

        line_n++;

        if(line.empty()){
            continue;
            }

        string bin_name, src_name, ver, rel;
        size_t index  = string::npos;
        size_t column = string::npos;
        bool skip = false;
        bool skipped_desc = false;

        // Defining a file format (altlinux bin.list or simple {bin, src} pairs)
        if(SRC_NAME_COL == string::npos){
            SRC_NAME_COL = SrcNamePos(line);
            }

        // For all words of a line
        for(;;){
            column++;
            index = GetWordFromLine(line, word, index + 1);
            // If new word has not been found in a line
            if(index == string::npos || word.empty()){
                break;
                }
            // Processing a bin package name
            if(column == BIN_NAME_COL){
                // Skipping some of the binary packages
                if(GotUnwantedSuffix(word)){
                    skipped_desc = true;
                    }
                bin_name = word;
                }
            // Processing a src package name of a bin package
            else if(column == SRC_NAME_COL){
                // Not going any further before we got a bin package name
                if(bin_name.empty()){
                    continue;
                    }
                // Skipping repeated entries
                if(src.find(bin_name) != src.end()){
                    skip = true;
                    break;
                    }
                // Determining a src name, a version and a release
                if(!DecomposeRPM(word, src_name, ver, rel)){
                    break;
                    }
                // Skipping excluded src
                if(excluded_src.find(src_name) != excluded_src.end()){
                    skip = true;
                    break;
                    }
                }
            }

        if(skip){
            continue;
            }

        if(bin_name.empty() || ver.empty() || rel.empty() || src_name.empty()){
            err = "Format violation in line " + to_string(line_n);
            return false;
            }

        // Adding a formed package description
        AddPackageDesc(src_name, bin_name, ver, rel, skipped_desc);
        }

    RestoreFromSkipped();

    return true;
    }

#undef BIN_NAME_COL

//-----------------------------------------------------------------------------
// Import an ACL of this branch

bool TBranch::ImportACL(const TFileInMem &file, const set<string> &excluded_src,
        string &err){

    string line;
    istringstream iss(file.data);

    // Loop through all lines of the file
    while(getline(iss,line)){

        string word;
        string src_name;
        size_t start_pos = string::npos;

        // For all words of a line
        for(;;){
            start_pos = GetWordFromLine(line, word, start_pos + 1);
            // If new word has not been found in a line
            if(start_pos == string::npos || word.empty()){
                break;
                }
            // Src name is a first word in a line
            if(src_name.empty()){
                src_name = word;
                // Skipping excluded src
                if(excluded_src.find(src_name) != excluded_src.end()){
                    break;
                    }
                continue;
                }
            // If it's not a src name than it's a name of a maintainer
            acl[src_name].push_back(word);
            }
        }

    return true;
    }

//------------------------------------------------------------------------------
// Get a version part from a ver-rel string

auto GetVer(const string &ver_rel){

    string ver;
    const size_t epoch_index = ver_rel.find(':');
    const size_t rel_index = ver_rel.rfind('-');

    if(rel_index != string::npos &&
            rel_index != 0 && rel_index != (ver_rel.size() - 1) &&
            (epoch_index == string::npos || epoch_index < rel_index)){
        const size_t ver_index = (epoch_index == string::npos) ?
            0 : (epoch_index + 1);
        ver = ver_rel.substr(ver_index, rel_index - ver_index);
        }

    return ver;
    }

//------------------------------------------------------------------------------

#define TIMELINE_TIMESTAMP  0   // Numbers of considered columns...
#define TIMELINE_SUBTASK    2   // ...in timeline files with timestamp...
#define TIMELINE_VER        3   // ...subtask num and ver-rel respectively

// Regex for checking if this is a 'qa' release
const regex qa_release_check("\\.qa[0-9]+$");

// Import timeline of a particular package
bool TBranch::ImportTimeLine(const TFileInMem &file,
        const set<string> &excluded_src, string &err){

    err.clear();

    // Form an error message about a wrong file path and return a true value
    auto wrong_path = [&](void){
        err = "Wrong path of the \"d-t-s-evr.list\" type of file "
            "\"" + file.fname + "\"";
        return true;
        };
    // - nested func

    // Determining a src name from an import file path
    const size_t i = file.fname.rfind("/");
    if(i == string::npos){
        return !wrong_path();
        }
    const string src_name = StripPath(file.fname.substr(0, i));
    if(src_name.empty()){
        return !wrong_path();
        }

    // Skipping excluded src
    if(excluded_src.find(src_name) != excluded_src.end()){
        return true;
        }

    int line_n = 0;
    string line;
    set<string> skipped_versions;
    istringstream iss(file.data);

    // Loop by all lines of the file
    while(getline(iss,line)){

        line_n++;

        if(line.empty()){
            continue;
            }

        string word, ver;
        unsigned timestamp = 0;
        size_t start_pos = string::npos;
        size_t column = string::npos;
        bool skip = false;

        // For all words of the line
        for(;;){
            column++;
            start_pos = GetWordFromLine(line, word, start_pos + 1);
            if(start_pos == string::npos || word.empty()){
                break;
                }
            if(column == TIMELINE_TIMESTAMP){
                if(word.front() == '-'){
                    break;
                    }
                try{
                    timestamp = stoi(word);
                    }
                catch(...){
                    timestamp = 0;
                    break;
                    }
                }
            else if(column == TIMELINE_SUBTASK){
                if(word == "-"){
                    skip = true;
                    }
                }
            else if(column == TIMELINE_VER){
                if(word == "-"){
                    skip = true;
                    break;
                    }
                ver = GetVer(word);
                if(skip || regex_search(word, qa_release_check)){
                    skip = true;
                    if(!ver.empty()){
                        skipped_versions.insert(ver);
                        }
                    break;
                    }
                }
            }

        if(skip){
            continue;
            }

        if(timestamp == 0 || ver.empty()){
            err = "Format violation in line " + to_string(line_n) + " of " +
                "\"" + src_name + "\" timeline";
            return false;
            }

        // Checking if there is an earlier timestamp for this version
        const auto src_it = timelines.find(src_name);
        if(src_it != timelines.end()){
            const auto &versions = src_it->second;
            const auto ver_it = versions.find(ver);
            if(ver_it != versions.end()){
                const auto &saved_timestamp = ver_it->second;
                if(timestamp >= saved_timestamp){
                    goto NextLine;
                    }
                }
            }

        // Checking if this version was skipped for some reason
        if(skipped_versions.find(ver) != skipped_versions.end()){
            goto NextLine;
            }

        timelines[src_name][ver] = timestamp;
        NextLine:
        ;
        }

    return true;
    }

#undef TIMELINE_TIMESTAMP
#undef TIMELINE_SUBTASK
#undef TIMELINE_VER

//-----------------------------------------------------------------------------
// Find relative packages

bool TBranch::DetectRelatives(void){

    printer.LineBegin("Detecting relatives for " + name + " src packages");

    const auto [counter, err] = helpers::DetectRelatives(src, PACK_RELATIVES);
    if(!err.empty()){
        printer.Err(err);
        return false;
        }

    printer.Success("Got " + to_string(counter) + " sets of related packages");
    return true;
    }

//-----------------------------------------------------------------------------
// Erase data related to packages that have given src names

void TBranch::Erase(const set<string> &src_names){

    for(const auto &src_name : src_names){
        src.erase(src_name);
        bin.erase(src_name);
        acl.erase(src_name);
        timelines.erase(src_name);
        }
    }

//-----------------------------------------------------------------------------
