/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2022 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef     cve_import_listsH
#define     cve_import_listsH

#include    <map>

#include    "main.h"

#define     PACK_DESC_SIZE  6   // Number of fields of package description
#define     PACK_VER        0   // Id of src-map-element with version
#define     PACK_REL        1   // Id of src-map-element with release
#define     PACK_FIXED_CVE  2   // Year and number of fixed CVE vulnerabilities
#define     PACK_FIXED_BDU  3   // Year and number of fixed BDU vulnerabilities
#define     PACK_URL        4   // Package URL
#define     PACK_RELATIVES  5   // Id of src-map-element with relatives

//-----------------------------------------------------------------------------
// Description of RPM repository

class TBranch{

    private:

        string name;        // Name of the branch
        string base_dir;    // Custom base dir stated in conf (/space/ALT/<name>/ by default)
        string bin_list;    // Custom bin_list path stated in conf
        string srpms_dir;   // Custom SRPMS path stated in conf
        vector<string> rpms_dirs;          // Paths of directories that contain RPM packages
        bool is_master_branch = false;     // Whether this is a master branch
        bool is_ghost_branch = false;      // Whether this is a master branch that will not be fully processed
        bool is_local_sys = false;         // Whether these are local system packages
        map<string, map<int, string>> src; // Descriptions of src packages
        map<string, map<int, string>> src_skipped;
        map<string, set<string>> bin;      // Bin packages of src packages
        map<string, set<string>> bin_skipped;
        map<string, vector<string>> acl;   // acl/list.packages.<name>
        map<string, map<string, unsigned>> timelines; // Timelines of src packages

        // Form a base path of this branch
        string BasePath(const string &alt_mnt_point);
        // Form full path for a given subpath inside a base directory
        string DirInsideBase(const string &alt_mnt_point, const string &subpath);

        // Form a full file path of a bin list of this branch
        string BinListPath(const string &alt_mnt_point, string &err);
        // Form full path of a directory that contain src packages
        string SRPMSDirPath(const string &alt_mnt_point, string &err);
        // Format paths of directories that contain bin packages
        vector<string> RPMSDirPaths(const string &alt_mnt_point, string &err);
        // Form a full file path of an acl file of this branch
        string ACLPath(const string &base_dir, string &err);
        // Form a vector of full file paths of timeline files of this branch
        vector<string> TimeLinesPaths(const string &base_dir, string &err);

        // Add formed package description
        void AddPackageDesc(const string &src_name, const string &bin_name,
                const string &ver, const string &rel, const bool skipped_desc);
        // Restore records for those src packages whose bin packages all have
        // unwanted suffixes
        void RestoreFromSkipped(void);

        // Import a bin list of this branch
        bool ImportBinList(const TFileInMem &file,
                const set<string> &exclued_src, string &err);
        // Form a list of local system packages using RPM
        bool LocalSysBinList(
                const set<string> &excluded_src, string &err);
        // Import an ACL of this branch
        bool ImportACL(const TFileInMem &file,
                const set<string> &exclued_src, string &err);
        // Import timeline of a particular package
        bool ImportTimeLine(const TFileInMem &file,
                const set<string> &exclued_src, string &err);

    public:

        // Data access
        const string& Name(void) const{
            return name;
            }
        const string& BaseDir(void) const{
            return base_dir;
            }
        bool IsMasterBranch(void) const{
            return is_master_branch;
            }
        bool IsGhostBranch(void) const{
            return is_ghost_branch;
            }
        bool IsLocalSys(void) const{
            return is_local_sys;
            }
        const map<string, map<int, string>>& Src(void) const{
            return src;
            }
        const map<string, set<string>>& Bin(void) const{
            return bin;
            }
        const map<string, map<string, unsigned>>& TimeLines(void) const{
            return timelines;
            }
        const map<string, vector<string>>& ACL(void) const{
            return acl;
            }

        // Modifiers
        void WriteName(const string &n){
            name = n;
            }
        void WriteBaseDir(const string &s){
            base_dir = s;
            }
        void WriteBinList(const string &s){
            bin_list = s;
            }
        void WriteSRPMSDir(const string &s){
            srpms_dir = s;
            }
        void WriteRPMSDirs(const string &s);
        void MakeMaster(void){
            is_master_branch = true;
            }
        void MakeGhost(void){
            is_master_branch = true;
            is_ghost_branch = true;
            }
        void MakeLocal(void){
            is_local_sys = true;
            }

        // Erase data related to packages that have given src names
        void Erase(const set<string> &src_names);

        // Import bin list or acl
        bool Import(const string &target, const string &base_dir,
                const set<string> &excluded_src);
        // Get IDs of fixed vulnerabilities and URLs from src or bin packages,
        // return a set of src names of missing packages and a flag that is
        // false in case of a critical error
        pair<set<string>, bool> ImportPackagesMetadata(
                const string &alt_mnt_point, const set<string> &excluded_src);

        // Find relative packages
        bool DetectRelatives(void);

        // Clear all data fields
        void Clear(void){
            name.clear();
            bin_list.clear();
            base_dir.clear();
            is_master_branch = false;
            src.clear();
            bin.clear();
            acl.clear();
            timelines.clear();
            }
    };

//-----------------------------------------------------------------------------

#endif
