/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2022 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    "main.h"
#include    "importer.h"

TRunMode run_mode;
TPrinter printer;

//-----------------------------------------------------------------------------
// Display help notes

void PrintHelp(void){

    const string exe = "cve-import ";

    cout << exe << "is an utility used to import lists of packages of examined" << endl
        << "repos, various lists with descriptions of vulnerabilities (in JSON" << endl
        << "and XML format) and a CPE dictionary into VUL DB." << endl
        << endl
        << "Import all available data:" << endl
        << "\t" << exe << "--everything" << endl
        << endl
        << "Import packages info (bin lists, acls, timelines):" << endl
        << "\t" << exe << "--packages [branch_i .. branch_j]" << endl
        << endl
        << "Import FSTEC vulnerabilities list:" << endl
        << "\t" << exe << "--fstec [xml_abs_path_i .. xml_abs_path_j]" << endl
        << endl
        << "Import NVD CVE list:" << endl
        << "\t" << exe << "--nvd [json_abs_path_i .. json_abs_path_j]" << endl
        << endl
        << "Import CPE dictionary:" << endl
        << "\t" << exe << "--cpe [xml_abs_path_i .. xml_abs_path_j]" << endl
        << endl
        << "Show this help message and exit:" << endl
        << "\t" << exe << "--help" << endl
        << endl;
    }

//-----------------------------------------------------------------------------
// Parse given command line parameters

pair<string, set<string>> ParseParams(const vector<string> &params){

    string target;
    set<string> specification;

    const set<string> available_targets =
        {"everything", "packages", TARGET_FSTEC, TARGET_NVD, TARGET_CPE};

    // Check that given target is one of the available targets
    auto is_available_target = [&](const string &t){
        return available_targets.find(t) != available_targets.end();
        };
    // - nested func

    for(const auto &param : params){

        if(param.empty()){
            continue;
            }

        // If this is not a flag
        if(param.find("--") != 0){
            if(target.empty() || target == "everything"){
                return {"", {}};
                }
            specification.insert(param);
            continue;
            }

        // This is some kind of a flag
        const auto fp = param.substr(2);

        if(fp == "help"){
            return {"help", {}};
            }

        if(fp == "debug"){
            run_mode.WriteFlag(DEBUG_FLAG, true);
            }
        else if(fp == "update"){
            run_mode.WriteFlag(UPDATE_FLAG, true);
            }
        else if(fp == "plain"){
            printer.Plain(true);
            }
        else if(fp == "silent"){
            printer.Silent(true);
            }
        else if(is_available_target(fp)){
            if(!target.empty()){
                printer.Err("Import of more than one specified target is not "
                    "supported; Use '--everything' flag to import all "
                    "available data (no additional parameters are required)");
                exit(1);
                }
            target = fp;
            }
        else{
            return {"", {}};
            }
        }

    return {target, specification};
    }

//-----------------------------------------------------------------------------

int main(int argc, char *argv[]){

    // Main executive unit
    TImporter importer;

    // Converting mass of mass of chars to vector of strings
    vector<string> params;
    for(int i = 1; i < argc; i++){
        params.push_back(argv[i]);
        }

    // Parsing the arguments
    const auto [target, specification] = ParseParams(params);

    if(target == "help" || target.empty()){
        PrintHelp();
        return target.empty() ? 0 : 1;
        }

    // Configuring the program
    if(!importer.Configure()){
        return 1;
        }

    if(target == "everything"){
        // Recreating the DB and importing everything that could be imported
        if(!importer.PrepareDB()){
            return 1;
            }
        return 0;
        }

    // Specific target is given, need to recreate a target table
    if(!importer.CreateTargetTable(target)){
        return 1;
        }

    if(target == "packages"){
        // Importing lists of packages for all considered branches
        if(!importer.ImportPackages(specification)){
            return 1;
            }
        }
    else{
        const bool force_clean = !specification.empty();
        // Importing lists of vulnerabilities or CPE dictionary
        if(!importer.ImportXmlOrJson(target, specification, force_clean)){
            return 1;
            }
        }

    return 0;
    }

//-----------------------------------------------------------------------------
