/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2022 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    <sstream>

#include    "related_package_names.h"

namespace helpers{

//-----------------------------------------------------------------------------
// Detect a special package name prefix in a given package name and return
// a length of a found prefix and a group id of a found prefix

#define GROUP_NONE  0
#define GROUP_LIB   1

pair<int, int> DetectSpecialPrefix(const string &name,
        const map<int, set<string>> &package_groups){

    for(const auto &[group_id, related_prefixes] : package_groups){
        for(const auto &prefix : related_prefixes){
            const size_t l = prefix.size();
            if(name.size() > l && name.compare(0, l, prefix) == 0){
                return {l, group_id};
                }
            }
        }

    return {-1, GROUP_NONE};
    };

//-----------------------------------------------------------------------------
// Produce a framed package name by marking a presence of a special prefix
// (second element of a returned pair) and trimming it, then right trimming
// non-alpha symbols, stripping delims and converting the rest to lowercase
// (that will be a first element of a returned pair)

const set<char> package_name_delims = {'.', '-', '_'};

pair<string, int> FramedPackageName(const string &name,
        const map<int, set<string>> &package_groups){

    string cname;
    const auto [prefix_length, group_id] = \
        DetectSpecialPrefix(name, package_groups);
    
    // Trimming and converting
    for(auto c = name.rbegin(); c != name.rend(); c++){
        if(group_id){
            const int n_chars_left = name.rend() - c;
            if(n_chars_left <= prefix_length){
                break;
                }
            }
        if(cname.empty()){
            if(!isalpha(*c)){
                continue;
                }
            }
        else{
            if(package_name_delims.find(*c) != package_name_delims.end()){
                continue;
                }
            }
        cname += tolower(*c);
        }

    reverse(cname.begin(), cname.end());

    return {cname, group_id};
    }

//-----------------------------------------------------------------------------
// Add a specified package name to a specified string of package names
// separated by spaces

void AppendRelative(const string &new_relative, string &relatives){

    if(!new_relative.empty()){
        if(!relatives.empty()){
            relatives += " ";
            }
        relatives += new_relative;
        }
    }

//------------------------------------------------------------------------------
// Read groups of packages from the "groups" file

#define GROUPS_COL_Q                6
#define GROUPS_COL_PACKAGE_PREFIXES 1

pair<const map<int, set<string>>, const string> GetPackageGroups(void){

    if(!CheckFile(GROUPS)){
        return {{}, "There is no \"" GROUPS "\" file"};
        }

    TFileInMem file;

    if(!file.ReadFileToMem(GROUPS)){
        return {{}, "Can't read \"" GROUPS "\" file"};
        }

    map<int, set<string>> package_groups = {{GROUP_LIB, {"lib"}}};
    string line;
    int line_n = 0;
    bool header = true;
    istringstream iss(file.data);

    // For all lines of a file that contains package/product groups
    while(getline(iss, line)){

        line_n++;
        Trim(line);
        if(line.empty()){
            continue;
            }

        if(header){
            header = false;
            continue;
            }

        auto columns = Split(line, ",");
        if(columns.size() != GROUPS_COL_Q){
            return {{}, FormatViolation(GROUPS, line_n)};
            }

        auto &prefixes_col = columns[GROUPS_COL_PACKAGE_PREFIXES];
        Trim(prefixes_col);

        const auto prefixes = Split(prefixes_col);
        if(prefixes.size() < 2){
            continue;
            }

        for(const auto &prefix : prefixes){
            package_groups[GROUP_LIB + line_n].insert(prefix + "-");
            }
        }

    return {package_groups, string()};
    }

#undef GROUPS_COL_Q
#undef GROUPS_COL_PREFIXES

//-----------------------------------------------------------------------------
// Detect related package names among given src packages represented with a map
// {<package_name> : <package_description>}, return a number of related
// package names

pair<unsigned, string> DetectRelatives(map<string, map<int, string>> &src,
        unsigned relatives_col){

    const auto [package_groups, err] = GetPackageGroups();

    if(!err.empty()){
        return {0, err};
        }

    map<string, pair<string, int>> framed_names;

    // Framing the package names    
    for(const auto &[src_package_name, _ ] : src){
        framed_names[src_package_name] = \
            FramedPackageName(src_package_name, package_groups);
        }

    // For some of the names relatives should not be considered
    const set<string> special_cases = {"firmware"};

    unsigned counter = 0;

    // For every package...
    for(auto it_a = src.begin(); it_a != src.end(); it_a++){

        auto &[package_name_a, package_desc_a] = *it_a;
        const auto &framed_name_a = framed_names[package_name_a];
        const auto &[stripped_name_a, group_a] = framed_name_a;

        if(special_cases.find(stripped_name_a) != special_cases.end()){
            continue;
            }

        // ...searching for related packages
        for(auto it_b = it_a; it_b != src.end(); it_b++){

            if(it_b == it_a){
                continue;
                }

            auto &[package_name_b, package_desc_b] = *it_b;
            const auto &framed_name_b = framed_names[package_name_b];
            const auto &[stripped_name_b, group_b] = framed_name_b;

            const bool divs_match =
                (group_a == group_b) ||
                (group_a == GROUP_LIB && group_b == GROUP_NONE) ||
                (group_b == GROUP_LIB && group_a == GROUP_NONE);

            if(divs_match && stripped_name_a == stripped_name_b){

                if(package_desc_a[relatives_col].empty() &&
                        package_desc_b[relatives_col].empty()){
                    counter++;
                    }

                AppendRelative(package_name_b, package_desc_a[relatives_col]);
                AppendRelative(package_name_a, package_desc_b[relatives_col]);
                }
            }
        }

    return {counter, string()};
    }

//-----------------------------------------------------------------------------

#undef GROUP_NONE
#undef GROUP_LIB

} // namespace helpers
