/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2022 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    <sstream>
#include    <filesystem>

#include    "rpm.h"
#include    "libcve-manager/src/vul.h"

namespace fs = filesystem;

//-----------------------------------------------------------------------------
// Run given command and get a pair of return code and command output

pair<int, string> Exec(const string &cmd){

    auto pipe = popen(cmd.c_str(), "r");

    if(!pipe){
        return pair<int, string>(-1, "Can't create pipe");
        }

    string stdout;
    array<char, 128> buffer;

    while(fgets(buffer.data(), buffer.size(), pipe) != NULL){
        stdout += buffer.data();
        }

    const int returncode = pclose(pipe);

    return pair<int, string>(returncode, stdout);
    }

//-----------------------------------------------------------------------------
// Get a map of all installed packages {<src_package>: {<bin_packages>}}

pair<TPackages, string> GetSysPackages(void){

    TPackages packages;

    // Requesting a list of all install packages
    const auto &[ret_code, ret_out] = Exec("rpm -qa");

    if(ret_code == -1){
        const string err = "Can't get a list of installed RPM packages";
        return {packages, err};
        }

    TBinPackages bin_packages;
    string bin_package;
    istringstream iss(ret_out);

    while(getline(iss, bin_package)){
        if(!bin_package.empty()){
            bin_packages.insert(bin_package);
            }
        }

    // Requesting SRPM name of every installed bin package
    for(const auto &bin_package : bin_packages){

        const auto &[ret_code, src_package] =
            Exec("rpm -q --qf \"%{SOURCERPM}\" " + bin_package);

        if(ret_code == -1 || src_package.empty()){
            const string err =
                "Can't get SRPM name of the \"" + bin_package + "\" package";
            return {packages, err};
            }

        packages[src_package].insert(bin_package);
        }

    return {packages, ""};
    }

//-----------------------------------------------------------------------------

TRpmInterface::TRpmInterface(const string &target, const bool is_local_sys){

    if(target == "fixes"){

        PostProcess = GetVulIDs;

        RequestRPM = is_local_sys ?
            [](const string &path){
                return Exec("rpm -q --changelog " + path);} :
            [](const string &path){
                return Exec("rpm -qp --changelog " + path);};

        AppendData = [](map<int, string> &src_desc, const set<string> &data){
            map<string, vector<string>> vul_ids;
            for(const auto &vul_id : data){
                if(vul_id.size() < 13){
                    continue;
                    }
                const string vul_marker = vul_id.substr(0, 3);
                const string vul_year_and_num = vul_id.substr(4);
                vul_ids[vul_marker].push_back(vul_year_and_num);
                }
            for(auto &[marker, index] : vul_markers){
                src_desc[index] = Join(vul_ids[marker], " ");
                }
            };
        }

    else if(target == "url"){

        PostProcess = [](const string &byte_array){
            set<string> res_set;
            string buffer;
            stringstream ss(byte_array, ios_base::in);
            while(getline(ss, buffer)){
                if(CheckSymbols(buffer, spec_symbols)){
                    buffer = Esc(buffer, spec_symbols);
                    }
                res_set.insert(buffer);
                }
            return pair<set<string>, string>(res_set, "");
            };

        RequestRPM = is_local_sys ?
            [](const string &path){
                return Exec("rpm -q --qf [%{url}\\\\n] " + path);} :
            [](const string &path){
                return Exec("rpm -qp --qf [%{url}\\\\n] " + path);};

        AppendData = [](map<int, string> &src_desc, const set<string> &data){
            auto &url = src_desc[PACK_URL];
            for(const auto &el : data){
                if(el.empty()){
                    continue;
                    }
                if(!url.empty()){
                    url += ",";
                    }
                url += el;
                }
            };
        }

    else{

        PostProcess = [](const string &s){
            return pair<set<string>, string>();
            };

        RequestRPM = [](const string &s){
            return pair<int, string>(-1, "");
            };

        AppendData = [](map<int, string> &src_desc, const set<string> &data){};
        }
    }

//-----------------------------------------------------------------------------
// Form a vector of names of all files of a specified directory, second value
// is true if operation is successful and is false otherwise

pair<vector<string>, bool> ListDir(const string &dir_path){

    if(!fs::exists(dir_path)){
        return {{}, false};
        }

    fs::directory_entry dir(dir_path);
    if(!dir.is_directory()){
        return {{}, false};
        }

    vector<string> paths;

    for(const auto &entry : fs::directory_iterator(dir)){
        if(!entry.is_regular_file()){
            continue;
            }
        paths.push_back(entry.path().filename());
        }

    return {paths, true};
    }

//-----------------------------------------------------------------------------
// Search for an rpm-filename of a specified package in a vector of file names,
// return an empty string if can't be found

string PackageFileName(const vector<string> &file_names,
        const string &package_name){

    const size_t l = package_name.size();

    // Iterating over specified directory
    for(const auto &file_name : file_names){
        if(file_name.size() > (l + 4) &&
                file_name.substr(0, l) == package_name &&
                file_name.substr(file_name.size() - 4) == ".rpm"){
            return file_name;
            }
        }

    return "";
    }

//-----------------------------------------------------------------------------
// Use given mediators to get IDs of fixed vulnerabilities and an URL of
// a package that has a specified file path, return a success flag

bool GetPackageMetadata(const map<string, TRpmInterface> &mediators,
        const string &path, const bool is_local_sys, map<int, string> &src_desc,
        map<string, int> &success_count, vector<string> &warn){

    // Check existence of a package
    if(!is_local_sys && !CheckFile(path)){
        warn.push_back("There is no \"" + path + "\" package");
        return false;
        }

    for(const auto &[target, mediator] : mediators){

        const auto &[ret_code, ret_out] = mediator.RequestRPM(path);

        pair<set<string>, string> ret_processed;

        if(ret_code == -1){
            warn.push_back("Can't get " + target + " from " + path);
            return false;
            }

        const auto [ret_set, current_warn] = mediator.PostProcess(ret_out);

        if(!current_warn.empty()){
            warn.push_back(path + " -- " + current_warn);
            }

        // If command succeeded and there is some output string
        if(ret_code == 0 && !ret_set.empty()){
            success_count[target] += 1;
            mediator.AppendData(src_desc, ret_set);
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Get IDs of fixed vulnerabilities stated in changelog of packages and get
// URLs from source or binary rpm packages of a given directories, insert it
// into "src" map and return a {<target>: <count of successful operations>} map
// and a set of missing src package names, push all warnings into "warn" vector

pair<map<string, int>, set<string>> GetMetadata(
        const string &srpms_dir,
        const vector<string> &rpms_dirs,
        map<string, map<int, string>> &src,
        const map<string, set<string>> &bin,
        const bool is_local_sys,
        const set<string> &excluded_src,
        vector<string> &warn){

    map<string, int> success_count;
    const array<string, 2> targets = {"fixes", "url"};
    map<string, TRpmInterface> mediators;

    for(auto &target : targets){
        mediators[target] = TRpmInterface(target, is_local_sys);
        success_count[target] = 0;
        }

    if(!is_local_sys && srpms_dir.empty() && rpms_dirs.empty()){
        warn.push_back("Both \"srpms_dir\" and \"rpms_dirs\" are not assigned");
        return {{}, Keys(src)};
        }

    set<string> missing_src;
    map<string, vector<string>> rpm_package_names;

    for(const auto &rpms_dir : rpms_dirs){
        const auto [file_names, ok] = ListDir(rpms_dir);
        if(!ok){
            warn.push_back("Can't list \"" + rpms_dir + "\" directory");
            return {{}, Keys(src)};
            }
        rpm_package_names[rpms_dir] = file_names;
        }

    // Getting rpm metadata from all packages or only from specified ones
    for(auto &[src_name, src_desc] : src){

        if(excluded_src.find(src_name) != excluded_src.end()){
            continue;
            }

        const auto &src_ver = src_desc.at(PACK_VER);
        const auto &src_rel = src_desc.at(PACK_REL);
        vector<string> warn_for_this_src;

        // If this call concerns the local system
        if(is_local_sys){

            const auto bin_packages = bin.find(src_name);

            if(bin_packages == bin.end()){
                goto MissingSrc;
                }

            const auto &[_, bin_names] = *bin_packages;

            for(const auto &bin_name : bin_names){

                if(GetPackageMetadata(mediators, bin_name, is_local_sys,
                        src_desc, success_count, warn_for_this_src)){
                    goto NextSrc;
                    }
                }

            goto MissingSrc;
            }

        // This call does not concern local system, trying to use
        // source package first
        if(!srpms_dir.empty()){

            const string package_name = src_name + "-" + src_ver + "-" + src_rel;
            const string package_path = srpms_dir + package_name + ".src.rpm";

            if(GetPackageMetadata(mediators, package_path, is_local_sys,
                    src_desc, success_count, warn_for_this_src)){
                continue;
                }
            }

        // Source package is not found, trying to use binary packages
        for(const auto &rpms_dir : rpms_dirs){

            if(rpms_dir.empty()){
                continue;
                }

            const auto bin_names = GetVal(bin, src_name);

            if(bin_names.empty()){
                const string msg = "Can't find names of bin packages of the "
                    "\"" + src_name + "\" source package";
                warn_for_this_src.push_back(msg);
                break;
                }

            // For every binary package name of this source package
            for(const auto &bin_name : bin_names){

                const auto &package_names = rpm_package_names[rpms_dir];
                const string package_name = PackageFileName(package_names,
                    bin_name + "-" + src_ver + "-" + src_rel);

                if(package_name.empty()){
                    const string msg = "Can't find \"" + bin_name + "\" "
                        "rpm package in \"" + rpms_dir + "\" directory";
                    warn_for_this_src.push_back(msg);
                    continue;
                    }

                const string package_path = JoinFilePaths(rpms_dir, package_name);

                if(GetPackageMetadata(mediators, package_path, is_local_sys,
                        src_desc, success_count, warn_for_this_src)){
                    goto NextSrc;
                    }
                }
            }

        MissingSrc:

        missing_src.insert(src_name);

        for(const auto &w : warn_for_this_src){
            warn.push_back(w);
            }

        NextSrc:
        ;
        }

    return {success_count, missing_src};
    }

//-----------------------------------------------------------------------------
