/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2022 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    "tree.h"

//-----------------------------------------------------------------------------
// Helper functions

string CaseError(const int index, const string &target){
    return ("There is no case for a \"" + to_string(index) + "\" " +
        target + " branch mark");
    }

string TargetError(const string &target, const string &format){
    return ("Disallowed combination of import target (" + target + ") "
        "and format (" + format + ")");
    }

//-----------------------------------------------------------------------------
// Form branch drafts to select data from XML document

pair<string, vector<tree::TBranchDraft>> FormXmlBranchDrafts(
        const string &target, string &err){

    string root;
    vector<tree::TBranchDraft> branch_drafts;
    tree::TBranchDraft branch_draft;

    #define FSTEC_ROOT "vulnerabilities"
    #define FSTEC_BASE {"vul"}
    #define FSTEC_SOFT {"vulnerable_software", "soft"}
    #define FSTEC_OS   {"environment", "os"}

    if(target == TARGET_FSTEC){
        root = FSTEC_ROOT;
        for(int i = 0; i < FSTEC_VUL_Q; i++){
            branch_draft.Clear();
            branch_draft.Add(FSTEC_BASE);
            switch(i){
                case FSTEC_VUL_YEAR:
                    break;
                case FSTEC_VUL_ID:
                    branch_draft.Add({"identifier"});
                    break;
                case FSTEC_VUL_NAME:
                    branch_draft.Add({"name"});
                    break;
                case FSTEC_VUL_DESCRIPTION:
                    branch_draft.Add({"description"});
                    break;
                case FSTEC_VUL_SOFT_VENDOR:
                    branch_draft.Add(FSTEC_SOFT);
                    branch_draft.Add({"vendor"});
                    break;
                case FSTEC_VUL_SOFT_NAME:
                    branch_draft.Add(FSTEC_SOFT);
                    branch_draft.Add({"name"});
                    break;
                case FSTEC_VUL_SOFT_VER:
                    branch_draft.Add(FSTEC_SOFT);
                    branch_draft.Add({"version"});
                    break;
                case FSTEC_VUL_SOFT_TYPE:
                    branch_draft.Add(FSTEC_SOFT);
                    branch_draft.Add({"types", "type"});
                    break;
                case FSTEC_VUL_ENV_VENDOR:
                    branch_draft.Add(FSTEC_OS);
                    branch_draft.Add({"vendor"});
                    break;
                case FSTEC_VUL_ENV_NAME:
                    branch_draft.Add(FSTEC_OS);
                    branch_draft.Add({"name"});
                    break;
                case FSTEC_VUL_ENV_VER:
                    branch_draft.Add(FSTEC_OS);
                    branch_draft.Add({"version"});
                    break;
                case FSTEC_VUL_ENV_PLATFORM:
                    branch_draft.Add(FSTEC_OS);
                    branch_draft.Add({"platform"});
                    break;
                case FSTEC_VUL_CWE:
                    branch_draft.Add({"cwe", "identifier"});
                    break;
                case FSTEC_VUL_IDENTIFY_DATE:
                    branch_draft.Add({"identify_date"});
                    break;
                case FSTEC_VUL_CVSS2_SCORE:
                    branch_draft.Add({"cvss", "vector"}, "score");
                    break;
                case FSTEC_VUL_CVSS3_SCORE:
                    branch_draft.Add({"cvss3", "vector"}, "score");
                    break;
                case FSTEC_VUL_SEVERITY:
                    branch_draft.Add({"severity"});
                    break;
                case FSTEC_VUL_SOLUTION:
                    branch_draft.Add({"solution"});
                    break;
                case FSTEC_VUL_STATUS:
                    branch_draft.Add({"vul_status"});
                    break;
                case FSTEC_VUL_EXPLOIT_STATUS:
                    branch_draft.Add({"exploit_status"});
                    break;
                case FSTEC_VUL_FIX_STATUS:
                    branch_draft.Add({"fix_status"});
                    break;
                case FSTEC_VUL_SOURCES:
                    branch_draft.Add({"sources"});
                    break;
                case FSTEC_VUL_OTHER_IDS:
                    branch_draft.Add({"identifiers", "identifier"});
                    break;
                default:
                    branch_drafts.clear();
                    err = CaseError(i, target);
                    return {{}, {}};
                }
            branch_draft.Complete(i);
            branch_drafts.push_back(branch_draft);
            }
        }

    #undef FSTEC_ROOT
    #undef FSTEC_BASE
    #undef FSTEC_SOFT
    #undef FSTEC_OS

    #define CPE_ROOT "cpe-list"
    #define CPE_BASE {"cpe-item"}

    else if(target == TARGET_CPE){
        root = CPE_ROOT;
        for(int i = 0; i < CPE_Q; i++){
            branch_draft.Clear();
            branch_draft.Add(CPE_BASE);
            switch(i){
                case CPE_FULLNAME:
                    branch_draft.Add({"cpe-23:cpe23-item"}, "name");
                    break;
                case CPE_TITLE:
                    branch_draft.Add({"title"});
                    break;
                case CPE_REFS:
                    branch_draft.Add({"references", "reference"}, "href");
                    break;
                default:
                    branch_drafts.clear();
                    err = CaseError(i, target);
                    return {{}, {}};
                }
            branch_draft.Complete(i);
            branch_drafts.push_back(branch_draft);
            }
        }
    else{
        err = TargetError(target, "xml");
        }

    #undef CPE_ROOT
    #undef CPE_BASE

    return {root, branch_drafts};
    }

//-----------------------------------------------------------------------------
// Form branch drafts to select data from JSON document

pair<string, vector<tree::TBranchDraft>> FormJsonBranchDrafts(
        const string &target, string &err){

    string root;
    vector<tree::TBranchDraft> branch_drafts;
    tree::TBranchDraft branch_draft;

    #define NVD_ROOT      "CVE_Items"
    #define NVD_NODES     {"configurations", "nodes", "?children"}
    #define NVD_CPE_MATCH {"configurations", "nodes", "?children", "cpe_match"}
    #define NVD_REFERENCE_DATA {"cve", "references", "reference_data"}

    if(target == TARGET_NVD){
        root = NVD_ROOT;
        for(int i = 0; i < NVD_CVE_Q; i++){
            int master_branch = -1;
            bool enable_indexing = false;
            switch(i){
                case NVD_CVE_YEAR:
                    break;
                case NVD_CVE_ID:
                    branch_draft.Add({"cve", "CVE_data_meta", "ID"});
                    break;
                case NVD_CVE_SOFTWARE:
                    branch_draft.Add(NVD_CPE_MATCH);
                    branch_draft.Add({"cpe23Uri"});
                    break;
                case NVD_CVE_SOFTWARE_VULNERABLE:
                    branch_draft.Add(NVD_CPE_MATCH);
                    branch_draft.Add({"vulnerable"});
                    master_branch = NVD_CVE_SOFTWARE;
                    break;
                case NVD_CVE_SOFTWARE_START_EXCL:
                    branch_draft.Add(NVD_CPE_MATCH);
                    branch_draft.Add({"versionStartExcluding"});
                    master_branch = NVD_CVE_SOFTWARE;
                    break;
                case NVD_CVE_SOFTWARE_START_INCL:
                    branch_draft.Add(NVD_CPE_MATCH);
                    branch_draft.Add({"versionStartIncluding"});
                    master_branch = NVD_CVE_SOFTWARE;
                    break;
                case NVD_CVE_SOFTWARE_END_EXCL:
                    branch_draft.Add(NVD_CPE_MATCH);
                    branch_draft.Add({"versionEndExcluding"});
                    master_branch = NVD_CVE_SOFTWARE;
                    break;
                case NVD_CVE_SOFTWARE_END_INCL:
                    branch_draft.Add(NVD_CPE_MATCH);
                    branch_draft.Add({"versionEndIncluding"});
                    master_branch = NVD_CVE_SOFTWARE;
                    break;
                case NVD_CVE_SOFTWARE_LOGIC1:
                    branch_draft.Add({"configurations", "nodes",
                        "operator"});
                    master_branch = NVD_CVE_SOFTWARE;
                    enable_indexing = true;
                    break;
                case NVD_CVE_SOFTWARE_LOGIC2:
                    branch_draft.Add({"configurations", "nodes", "children",
                        "operator"});
                    master_branch = NVD_CVE_SOFTWARE;
                    enable_indexing = true;
                    break;
                case NVD_CVE_PUBLISHED:
                    branch_draft.Add({"publishedDate"});
                    break;
                case NVD_CVE_MODIFIED:
                    branch_draft.Add({"lastModifiedDate"});
                    break;
                case NVD_CVE_CVSS_SCORE_V2:
                    branch_draft.Add({"impact", "baseMetricV2", "cvssV2",
                        "baseScore"});
                    break;
                case NVD_CVE_CVSS_SCORE_V3:
                    branch_draft.Add({"impact", "baseMetricV3", "cvssV3",
                        "baseScore"});
                    break;
                case NVD_CVE_CWE_ID:
                    branch_draft.Add({"cve", "problemtype", "problemtype_data",
                        "description", "value"});
                    break;
                case NVD_CVE_REFS:
                    branch_draft.Add(NVD_REFERENCE_DATA);
                    branch_draft.Add({"url"});
                    break;
                case NVD_CVE_REFS_TAG:
                    branch_draft.Add(NVD_REFERENCE_DATA);
                    branch_draft.Add({"tags"});
                    master_branch = NVD_CVE_REFS;
                    break;
                case NVD_CVE_SUMMARY:
                    branch_draft.Add({"cve", "description", "description_data",
                        "value"});
                    break;
                default:
                    branch_drafts.clear();
                    err = CaseError(i, target);
                    return {{}, {}};
                }
            branch_draft.Complete(i, master_branch, enable_indexing);
            branch_drafts.push_back(branch_draft);
            branch_draft.Clear();
            }
        }
    else{
        err = TargetError(target, "json");
        }

    #undef NVD_ROOT
    #undef NVD_NODES
    #undef NVD_CPE_MATCH
    #undef NVD_REFERENCE_DATA

    return {root, branch_drafts};
    }

//-----------------------------------------------------------------------------
